import pandas as pd
import numpy as np
import numpy
import cv2
import scanpy as sc
import logging
import anndata

from spacemake.errors import SpacemakeError

logger_name = "spacemake.spatial"
logger = logging.getLogger(logger_name)


def create_spot_expression_img(
    adata: anndata.AnnData,
    binary: bool = True,
) -> (numpy.ndarray, numpy.ndarray):
    """
    Creates a binary/grayscale image based on spatial expression data.

    Given a spacemake processed AnnData object this function will create
    two images based on the spatial expression values, one grayscale and
    one binary (black and white). For each observation in the data
    (spot/spatial-unit) a circle will be drawn, where the diameter of the
    circle is equal to the spot_diameter_um  provided in the puck parameters
    for this sample. In the binary image each circle will be black and
    for the grayscale image the color of each circle will correspond to the
    UMI count of that spot. Both image returned are 8-bit.

    :param adata: A spacemake processed AnnData object.
    :type adata: anndata.AnnData
    :returns: A tuple of (spot_grayscale_img, spot_binary_img)
    :rtype: (numpy.ndarray, numpy.ndarray)
    """
    df = adata.obs.copy()
    x_pos_max = df.x_pos.max()
    y_pos_max = df.y_pos.max()
    coord_by_um = adata.uns["puck_variables"]["coord_by_um"]

    width = int(x_pos_max / coord_by_um)
    height = int(y_pos_max / coord_by_um)

    spot_diameter = int(adata.uns["puck_variables"]["spot_diameter_um"])

    width = width + spot_diameter
    height = height + spot_diameter

    spot_img = np.zeros((height, width), np.uint8)
    spot_img_bw = np.zeros((height, width), np.uint8)

    df.total_counts = df.total_counts * 255 / df.total_counts.max()

    df.x_pos = df.x_pos * height / x_pos_max + int(spot_diameter / 2)
    df.y_pos = df.y_pos * width / y_pos_max + int(spot_diameter / 2)

    for i, row in df.iterrows():
        color = int(row["total_counts"])

        # draw the circle
        cv2.circle(
            spot_img,
            (int(row["y_pos"]), int(row["x_pos"])),
            int(spot_diameter / 2),
            (color, color, color),
            -1,
        )

        # draw the circle
        cv2.circle(
            spot_img_bw,
            (int(row["y_pos"]), int(row["x_pos"])),
            int(spot_diameter / 2),
            (255, 255, 255),
            -1,
        )

    spot_img = 255 - spot_img
    spot_img_bw = 255 - spot_img_bw

    spot_img = cv2.resize(spot_img, (1000, 1000), cv2.INTER_AREA)
    spot_img_bw = cv2.resize(spot_img_bw, (1000, 1000), cv2.INTER_AREA)

    spot_img_bw[spot_img_bw < 255] = 0

    spot_img_cnt = np.where(spot_img_bw < 255)
    spot_img = spot_img[
        spot_img_cnt[0].min() : spot_img_cnt[0].max(),
        spot_img_cnt[1].min() : spot_img_cnt[1].max(),
    ]

    spot_img_bw = spot_img_bw[
        spot_img_cnt[0].min() : spot_img_cnt[0].max(),
        spot_img_cnt[1].min() : spot_img_cnt[1].max(),
    ]

    return spot_img, spot_img_bw


def create_aggregated_expression_img(
    adata: anndata.AnnData, binary_top_qth_percentile: int = 30
) -> (numpy.ndarray, numpy.ndarray):
    """
    Creates an aggregated binary/grayscale image based on spatial expression data.

    Similar to create_spot_expression_img. Rather than draw a circle for each
    spot, data will be aggregated into a 500 x 500 pixel image. For grayscale
    images the value of each pixel will be corresponding to the sum of UMI
    counts under that pixel. For the binary image, pixels will be filtered
    based on the binary_top_qth_percentile parameter. After aggregation only
    the binary_top_qth_percentile pixels (the top 30% of pixels based on
    expression value) will be set. After this step the image will be smoothed
    and pixels with enough set neighbors will also be set. Both images
    returned are 8-bit.


    :param adata: A spacemake processed AnnData object.
    :type adata: anndata.AnnData
    :param binary_top_qth_percentile:
    :type binary_top_qth_percentile: int
    :returns: A tuple of (aggregated_grayscale_img, aggregated_binary_img)
    :rtype: (numpy.ndarray, numpy.ndarray)
    """
    df = adata.obs.copy()
    # load the clusters in their places as per coordinates
    df.x_pos = df.x_pos.astype("int")
    df.y_pos = df.y_pos.astype("int")

    max_x = df.x_pos.max()
    max_y = df.y_pos.max()

    img = np.zeros((max_x + 1, max_y + 1), np.uint8)

    img[df.x_pos, df.y_pos] = df.total_counts

    if max_x <= 500 or max_y <= 500:
        raise SpacemakeError(
            "Underlying spatial coordinates cannot be aggregated\n"
            + "Only coordinates with maximum values for x_pos and y_pos "
            + "both higher than 500 can be aggregated\n"
            + "Use `spacemake spatial create_spot_expression_img` instead"
        )

    scale_f = int(img.shape[0] / 500)

    # creating a 500x500 pixel image
    img_scaled = np.add.reduceat(img, range(0, img.shape[0], scale_f))
    img_scaled = np.add.reduceat(img_scaled, range(0, img.shape[1], scale_f), axis=1)

    # get only top 30%, by default
    if binary_top_qth_percentile > 0 and binary_top_qth_percentile <= 100:
        filter_val = np.percentile(
            img_scaled.flatten(), 100 - binary_top_qth_percentile
        )
    else:
        raise ValueError(
            f"binary_top_qth_percentile has to be between 1 and 100."
            f"The value provided ({binary_top_qth_percentile}) is invalid"
        )

    img_scaled_bw = img_scaled.copy()

    img_scaled_bw[img_scaled_bw > filter_val] = 255
    img_scaled_bw[img_scaled_bw <= filter_val] = 0

    img_scaled_bw = img_scaled_bw.astype(np.uint8)

    # fill black
    img_scaled_bw = img_scaled_bw | fill_image(img_scaled_bw, 15, 5)
    # fill white
    img_scaled_bw = ~img_scaled_bw
    img_scaled_bw = img_scaled_bw | fill_image(img_scaled_bw, 21, 5)
    img_scaled_bw = ~img_scaled_bw
    # fill black

    img_scaled = img_scaled * 255 / img_scaled.max()
    img_scaled = img_scaled.astype(np.uint8)

    return ~img_scaled, ~img_scaled_bw


def fill_image(X, n_neighbors=3, infer_square_side=3):
    """fill_image.

    :param X:
    :param n_neighbors:
    :param infer_square_side:
    """
    # create a zero array with + 1 row/column at the start and at the end
    # completely white image
    width, height = X.shape
    Y = np.zeros((width + infer_square_side, height + infer_square_side), np.uint16)

    for i in range(infer_square_side):
        for j in range(infer_square_side):
            Y[i : (width + i), j : (height + j)] = (
                Y[i : (width + i), j : (height + j)] + X
            )

    # get the middle
    Y = Y[1 : (width + 1), 1 : (height + 1)] - X

    # get only the pixels which have at least n_neighbors black neighbors and set them black
    Y[np.where(Y < 255 * n_neighbors)] = 0
    Y[np.where(Y >= 255 * n_neighbors)] = 255
    Y = Y.astype(np.uint8)

    return Y


def fill_holes_by_neighbors(dat, dist=2):
    """fill_holes_by_neighbors.

    :param dat:
    :param dist:
    """
    top = np.zeros(dat.shape, dtype=np.uint8)
    bottom = np.zeros(dat.shape, dtype=np.uint8)
    left = np.zeros(dat.shape, dtype=np.uint8)
    right = np.zeros(dat.shape, dtype=np.uint8)
    h, w = dat.shape

    for i in range(dist):
        top[0 : (h - i), :] = top[0 : (h - i), :] | dat[i:h, :]
        bottom[i:h, :] = bottom[i:h, :] | dat[0 : (h - i), :]
        right[:, 0 : (w - i)] = right[:, 0 : (w - i)] | dat[:, i:w]
        left[:, i:w] = right[:, i:w] | dat[:, 0 : (w - i)]

    return top & bottom & left & right


def load_he_img(he_path, bw_threshold=None):
    """load_he_img.

    :param he_path:
    :param bw_threshold:
    """
    he = cv2.imread(he_path, cv2.IMREAD_COLOR)
    he_gray = cv2.cvtColor(he, cv2.COLOR_BGR2GRAY)

    # create binary image
    if bw_threshold is None:
        thresh, he_bw = cv2.threshold(
            he_gray, 127, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU
        )
    else:
        he_bw = cv2.threshold(he_gray, bw_threshold, 255, cv2.THRESH_BINARY)[1]

    return he, he_gray, he_bw


def align_he_img(he_path, expression_img, bw_threshold=None, use_bw=True):
    """align_he_img.

    :param he_path:
    :param expression_img:
    :param bw_threshold:
    :param use_bw:
    """
    he, he_gray, he_bw = load_he_img(he_path, bw_threshold=bw_threshold)

    he_orig = he.copy()

    # find the scale by which we resize the images
    expression_img_height, expression_img_width = expression_img.shape
    he_height, he_width = he_gray.shape

    max_zoom = 3

    height_ratio = he_height / (max_zoom * expression_img_height)
    width_ratio = he_width / (max_zoom * expression_img_width)

    resize_type = cv2.INTER_NEAREST if use_bw else cv2.INTER_AREA

    if height_ratio > 1.0 and width_ratio > 1.0:
        # we scale the he image
        if height_ratio > width_ratio:
            scale_f = 1.0 / width_ratio
        else:
            scale_f = 1.0 / height_ratio

        dim = (int(he_width * scale_f) + 1, int(he_height * scale_f) + 1)

        he = cv2.resize(he, dim, 0, 0, cv2.INTER_AREA)
        he_gray = cv2.resize(he_gray, dim, 0, 0, cv2.INTER_AREA)
        he_bw = cv2.resize(he_bw, dim, 0, 0, cv2.INTER_NEAREST)

    else:
        if height_ratio > width_ratio:
            scale_f = width_ratio
        else:
            scale_f = height_ratio

        expression_img = cv2.resize(
            expression_img,
            (int(expression_img_width * scale_f), int(expression_img_height * scale_f)),
            resize_type,
        )

    he_bw = ~he_bw

    for i in range(10):
        he_bw = fill_holes_by_neighbors(he_bw, int(max(he_bw.shape) / 200))

    he_bw = ~he_bw

    highest_cor = 0.0
    scale_dim = None
    align_res = None
    rotate = 0
    flip = False

    he_res = None

    for make_flip in [False, True]:
        he_copy = he_bw.copy() if use_bw else he_gray.copy()

        if make_flip:
            he_copy = cv2.flip(he_copy, 0)

        for rotate_n in [0, 1, 2, 3]:
            he_rotated = he_copy
            for r in range(rotate_n):
                he_rotated = cv2.rotate(he_rotated, cv2.ROTATE_90_CLOCKWISE).copy()

            for scale_x in np.linspace(1.0, 1.0 / max_zoom + 0.01, 50):
                scales_y = (
                    min(1.0, scale_x * 1.1),
                    max(1.0 / max_zoom + 0.01, scale_x * 0.9),
                )
                for scale_y in np.linspace(scales_y[0], scales_y[1], 10):
                    dim = (int(he_bw.shape[1] * scale_x), int(he_bw.shape[0] * scale_y))

                    he_scaled = cv2.resize(he_rotated, dim, 0, 0, resize_type)
                    m_res = cv2.matchTemplate(
                        he_scaled, expression_img, cv2.TM_CCOEFF_NORMED
                    )

                    if m_res.max() > highest_cor:
                        highest_cor = m_res.max()
                        scale_dim = dim
                        align_res = m_res
                        rotate = rotate_n
                        flip = make_flip

    # find the boundaries of the align
    min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(align_res)
    top_left = max_loc
    bottom_right = (
        top_left[0] + expression_img.shape[1],
        top_left[1] + expression_img.shape[0],
    )

    # expression_img[expression_img > 0] = 255
    expression_img_clr = cv2.cvtColor(expression_img, cv2.COLOR_GRAY2BGR)

    if flip:
        he = cv2.flip(he, 0)
        he_gray = cv2.flip(he_gray, 0)
        he_bw = cv2.flip(he_bw, 0)
        he_orig = cv2.flip(he_orig, 0)

    for r in range(rotate):
        he = cv2.rotate(he, cv2.ROTATE_90_CLOCKWISE)
        he_gray = cv2.rotate(he_gray, cv2.ROTATE_90_CLOCKWISE)
        he_bw = cv2.rotate(he_bw, cv2.ROTATE_90_CLOCKWISE)
        he_orig = cv2.rotate(he_orig, cv2.ROTATE_90_CLOCKWISE)

    he = cv2.resize(he, scale_dim, 0, 0, cv2.INTER_AREA)
    he_bw_res = cv2.resize(he_bw, scale_dim, 0, 0, cv2.INTER_NEAREST)
    he_gray_res = cv2.resize(he_gray, scale_dim, 0, 0, cv2.INTER_AREA)

    return highest_cor, he, he_orig, top_left, bottom_right


def align_he_spot_img(
    adata: anndata.AnnData, he_path: str
) -> (numpy.ndarray, numpy.ndarray):
    """
    Align a H&E image with a spacemake processed data, based on spot expression.

    Takes an H&E image and a spacemake processed dataset, and alignes the two,
    trying to find the most probably position of the spatial dataset within the
    H&E image. This function should be used for datasets with a spot_diameter_um
    higher than or equal to 10 microns, such as Visium or Slide-seq.

    :param adata: A spacemake processed AnnData object.
    :type adata: anndata.AnnData
    :param he_path: Path to the H&E image to be aligned.
    :type he_path: str
    :returns: A tuple of (original_he, aligned_he). original_he will contain the
        original H&E data together with the aligned spots and a blue rectangle
        showing the aligned region. aligned_he will contain the part of the
        H&E image aligned.
    :rtype: (numpy.ndarray, numpy.ndarray)
    """
    spot_img, spot_img_bw = create_spot_expression_img(adata)

    # we align the binary image
    spot_img = spot_img_bw

    # find the contour of the spot image
    spot_img_cnt = np.where(spot_img < 255)
    spot_img = spot_img[
        spot_img_cnt[0].min() : spot_img_cnt[0].max(),
        spot_img_cnt[1].min() : spot_img_cnt[1].max(),
    ]

    highest_cor, he_res, he_orig, tl, br = align_he_img(
        he_path, expression_img=spot_img, use_bw=True
    )

    he_res_ratio = he_res.shape[1] / he_res.shape[0]

    he_orig = cv2.resize(
        he_orig,
        (int(he_orig.shape[0] * he_res_ratio), he_orig.shape[0]),
        cv2.INTER_AREA,
    )

    horizontal_ratio = he_orig.shape[0] / he_res.shape[0]
    vertical_ratio = he_orig.shape[1] / he_res.shape[1]

    box_tl = int(tl[0] * horizontal_ratio), int(tl[1] * vertical_ratio)
    box_br = int(br[0] * horizontal_ratio), int(br[1] * vertical_ratio)

    spot_img_resized = cv2.resize(
        spot_img, (box_br[0] - box_tl[0], box_br[1] - box_tl[1]), cv2.INTER_NEAREST
    )

    align_bw = cv2.cvtColor(spot_img_resized, cv2.COLOR_GRAY2BGR)

    he_align = he_orig[box_tl[1] : box_br[1], box_tl[0] : box_br[0]].copy()
    he_orig[box_tl[1] : box_br[1], box_tl[0] : box_br[0]] &= align_bw

    cv2.rectangle(he_orig, box_tl, box_br, (180, 233, 86), 3)

    return he_orig, he_align


def align_he_aggregated_img(
    adata: anndata.AnnData,
    he_path: str,
    bw_threshold: int = 200,
    binary_top_qth_percentile: int = 30,
    box_size: float = 0.5,
) -> (numpy.ndarray, numpy.ndarray):
    """align_he_aggregated_img.
    Align a H&E image with a spacemake processed data, based on aggregated
    expression.

    Takes an H&E image and a spacemake processed dataset, and alignes the two,
    trying to find the most probably position of the spatial dataset within the
    H&E image. First the spatial data will be aggregated in a 500 x 500 pixel
    image, then the best align between the spatial expression image and the H&E
    will be found. This function should be used for datasets with a
    spot_diameter_um smaller than 10 microns, such as Visium or Slide-seq.

    :param adata: A spacemake processed AnnData object.
    :type adata: anndata.AnnData
    :param he_path: Path to the H&E image to be aligned.
    :type he_path: str
    :param bw_threshold:
    :type bw_threshold: int
    :param binary_top_qth_percentile:
    :type binary_top_qth_percentile: int
    :param box_size: By default not the whole aggregated image will be aligned,
        only the middle part (a box with 50% x 50% pixels, wrt the original size).
        This parameter controls the size of the aligning box.
    :type box_size: float
    :returns: A tuple of (original_he, aligned_he). original_he will contain the
        original H&E data together with the aggregated image overlayed, together
        with the aligned area and the aligned_box shown as blue rectangles.
        aligned_he will contain the part of the H&E image aligned.
    :rtype: (numpy.ndarray, numpy.ndarray)
    """
    img, img_bw = create_aggregated_expression_img(adata, binary_top_qth_percentile)

    h, w = img.shape

    bottom_right = (int(w * ((1 + box_size) / 2)), int(h * ((1 + box_size) / 2)))
    top_left = (int(w * ((1 - box_size) / 2)), int(h * ((1 - box_size) / 2)))

    img = img_bw[top_left[1] : bottom_right[1], top_left[0] : bottom_right[0]]

    highest_cor, he_res, he_orig, tl, br = align_he_img(
        he_path, expression_img=img, bw_threshold=bw_threshold, use_bw=True
    )

    align_w = br[0] - tl[0]
    align_h = br[1] - tl[1]
    multiplier = (1 - box_size) / (box_size * 2)
    box_tl = (int(tl[0] - align_w * multiplier), int(tl[1] - align_h * multiplier))
    # box_tl = h,w
    box_br = (int(br[0] + align_w * multiplier), int(br[1] + align_h * multiplier))

    he_res_ratio = he_res.shape[1] / he_res.shape[0]

    he_orig = cv2.resize(
        he_orig,
        (int(he_orig.shape[0] * he_res_ratio), he_orig.shape[0]),
        cv2.INTER_AREA,
    )

    horizontal_ratio = he_orig.shape[0] / he_res.shape[0]
    vertical_ratio = he_orig.shape[1] / he_res.shape[1]

    # scale boxes
    tl = int(tl[0] * horizontal_ratio), int(tl[1] * vertical_ratio)
    br = int(br[0] * horizontal_ratio), int(br[1] * vertical_ratio)
    box_tl = (int(box_tl[0] * horizontal_ratio), int(box_tl[1] * vertical_ratio))
    box_br = (int(box_br[0] * horizontal_ratio), int(box_br[1] * vertical_ratio))

    img_resized = cv2.resize(
        img_bw, (box_br[0] - box_tl[0], box_br[1] - box_tl[1]), cv2.INTER_NEAREST
    )

    align_bw = cv2.cvtColor(img_resized, cv2.COLOR_GRAY2BGR)

    he_align = he_orig[box_tl[1] : box_br[1], box_tl[0] : box_br[0]].copy()

    cv2.rectangle(he_orig, tl, br, (180, 233, 86, 255), 5)
    cv2.rectangle(he_orig, box_tl, box_br, (180, 233, 86, 255), 5)

    overlay = he_orig.copy()
    overlay[box_tl[1] : box_br[1], box_tl[0] : box_br[0]] &= align_bw

    cv2.addWeighted(overlay, 0.4, he_orig, 0.6, 0, he_orig)

    return he_orig, he_align


def attach_he_adata(
    adata: anndata.AnnData,
    aligned_he: numpy.ndarray,
    push_by_spot_diameter: bool = True,
    raw_aligned: bool = False,
) -> anndata.AnnData:
    """
    Attach the aligned H&E data to a spacemake processed AnnData object.

    :param adata: A spacemake processed AnnData object.
    :type adata: anndata.AnnData
    :param aligned_he: A spacemake aligned H&E image aligned either with the
        align_he_aggregated_img or the align_he_spot_img functions.
    :type aligned_he: numpy.ndarray
    :param push_by_spot_diameter: If true the aligned image will be pushed by
        one spot diameter. To be used only if a spot image was aligned.
        Defaults to True.
    :type push_by_spot_diameter: bool
    :param raw_aligned: If true it is assumed that the H&E image was aligned
        using the raw spacemake processed AnnData. If a spot image was aligned
        this should be set to False. Defaults to False.
    :type raw_aligned: bool
    :returns: An AnnData object with the attached H&E image. When used with
        scanpy's scanpy.pl.spatial, under the spatial data the H&E image will
        be shown.
    :rtype: anndata.AnnData
    """
    import math

    # get the width of the puck
    puck_width_um = adata.uns["puck_variables"]["width_um"]
    spot_diameter_um = adata.uns["puck_variables"]["spot_diameter_um"]

    if adata.uns["run_mode_variables"]["mesh_data"]:
        spot_diameter_um = (
            math.sqrt(3) * adata.uns["run_mode_variables"]["mesh_spot_diameter_um"]
        )

    # get the width in coordinates
    width_coord_max = (
        adata.uns["puck_variables"]["width_um"]
        * adata.uns["puck_variables"]["coord_by_um"]
    )
    height_coord_max = (
        adata.uns["puck_variables"]["height_um"]
        * adata.uns["puck_variables"]["coord_by_um"]
    )

    spot_distance_um = 1 / adata.uns["puck_variables"]["coord_by_um"]

    # get the width of the processed adata
    adata_width_um = spot_distance_um * (adata.obsm["spatial"].max(axis=0)[0])

    # rotate he to align with coordinate system of scanpy
    rotated_he = cv2.flip(cv2.rotate(aligned_he, cv2.ROTATE_90_CLOCKWISE), 1)

    h_px, w_px = rotated_he.shape[:2]

    px_per_um = w_px / adata_width_um

    # convert so that origin at 0, 0
    locations = adata.obsm["spatial"]

    if not raw_aligned:
        locations = locations - locations.min(axis=0)

    spot_diameter_px = spot_diameter_um * px_per_um

    if raw_aligned:

        locations = locations * [(w_px) / width_coord_max, (h_px) / height_coord_max]
    else:
        locations = locations * [
            (w_px) / locations.max(axis=0)[0],
            (h_px) / locations.max(axis=0)[1],
        ]

    if push_by_spot_diameter:
        locations = [spot_diameter_px, spot_diameter_px] + locations

    locations = locations.astype("int")

    adata.uns["spatial"] = {
        "img": {
            "images": {"hires": rotated_he},
            "scalefactors": {
                "tissue_hires_scalef": 1,
                "spot_diameter_fullres": spot_diameter_px,
            },
        }
    }
    adata.obsm["spatial"] = locations

    return adata



import logging

logger_name = "spacemake.preprocess.dge"
logger = logging.getLogger(logger_name)

def calculate_adata_metrics(adata, dge_summary_path=None, n_reads=None):
    import scanpy as sc
    import pandas as pd

    # calculate mitochondrial gene percentage
    adata.var["mt"] = (
        adata.var_names.str.startswith("Mt-")
        | adata.var_names.str.startswith("mt-")
        | adata.var_names.str.startswith("MT-")
    )

    sc.pp.calculate_qc_metrics(
        adata, qc_vars=["mt"], percent_top=None, log1p=False, inplace=True
    )

    add_reads = False
    if dge_summary_path is not None:
        dge_summary = pd.read_csv(
            dge_summary_path,
            skiprows=7,
            sep="\t",
            index_col="cell_bc",
            names=["cell_bc", "n_reads", "n_umi", "n_genes"],
        )

        adata.obs = pd.merge(
            adata.obs, dge_summary[["n_reads"]], left_index=True, right_index=True
        )

        add_reads = True

    if n_reads is not None:
        adata.obs["n_reads"] = n_reads
        add_reads = True

    if add_reads:
        adata.obs["reads_per_counts"] = adata.obs.n_reads / adata.obs.total_counts


def calculate_shannon_entropy_scompression(adata):
    import math
    import itertools
    import numpy as np
    from collections import Counter

    def compute_shannon_entropy(barcode):
        prob, length = Counter(barcode), np.float16(len(barcode))
        return np.float16(-sum(
            np.float16(count) / length * np.log2(np.float16(count) / length) for count in prob.values()
        ))

    def compute_string_compression(barcode):
        compressed_barcode = "".join(
            letter + str(len(list(group)))
            for letter, group in itertools.groupby(barcode)
        )
        return np.uint16(len(compressed_barcode))

    bc = adata.obs.index.to_numpy()
    bc_len = len(bc[0])
    theoretical_barcodes = np.random.choice(
        ["A", "C", "T", "G"], size=(bc.shape[0], bc_len)
    )

    adata.obs["exact_entropy"] = np.round(
        np.array([compute_shannon_entropy(cell_bc) for cell_bc in bc]), 2
    )
    adata.obs["theoretical_entropy"] = np.round(
        np.array(
            [compute_shannon_entropy(cell_bc) for cell_bc in theoretical_barcodes]
        ),
        2,
    )
    adata.obs["exact_compression"] = np.round(
        np.array([compute_string_compression(cell_bc) for cell_bc in bc]), 2
    )
    adata.obs["theoretical_compression"] = np.round(
        np.array(
            [compute_string_compression(cell_bc) for cell_bc in theoretical_barcodes]
        ),
        2,
    )


def dge_to_sparse_adata(dge_path, dge_summary_path):
    import anndata
    import numpy as np
    import gzip
    import pandas as pd
    from scipy.sparse import coo_matrix, hstack

    gene_names = []

    with gzip.open(dge_path, "rt") as dge:
        first_line = dge.readline().strip().split("\t")
        has_mt = False
        barcodes = first_line[1:]
        N_bc = len(barcodes)
        X = None

        # read DGE line by line
        # first row: contains CELL BARCODEs
        # each next row contains one gene name, and the counts of that gene
        for line in dge:
            vals = line.strip()
            _idx_tab = vals.index("\t")
            _gene_name = vals[:_idx_tab]
            gene_names.append(_gene_name)

            if _gene_name.lower().startswith("mt-"):
                has_mt = True
                
            # store counts as np.array
            _vals = np.fromstring(vals[_idx_tab:], dtype=np.int32, count=N_bc, sep='\t').flatten()
            _idx_nonzero = np.argwhere(_vals != 0).flatten()

            if len(_idx_nonzero) > 0:
                gene_sp = coo_matrix((_vals[_idx_nonzero].astype(np.int32), (_idx_nonzero, np.zeros(len(_idx_nonzero)))), shape=(N_bc, 1), dtype=np.int32)
            else:
                gene_sp = coo_matrix((N_bc, 1), dtype=np.int32)

            if X is None:
                 X = gene_sp
            else:
                 X = hstack([X, gene_sp])

        if X is None:
            X = coo_matrix((len(barcodes), 0), dtype=np.int32)
    
        if not has_mt:
            # ensure we have an entry for mitochondrial transcripts even if it's just all zeros
            print(
                "need to add mt-missing because no mitochondrial stuff was among the genes for annotation"
            )
            gene_names.append("mt-missing")
            X = hstack([X, np.zeros(X.shape[0])[:, None]])

        X = X.tocsr()
        X = X.astype(np.float32)
        adata = anndata.AnnData(
            X, obs=pd.DataFrame(index=barcodes), var=pd.DataFrame(index=gene_names)
        )

        # name the index
        adata.obs.index.name = "cell_bc"

        # attach metrics such as: total_counts, pct_mt_counts, etc
        # also attach n_genes, and calculate pcr
        calculate_adata_metrics(adata, dge_summary_path)

        # calculate per shannon_entropy and string_compression per bead
        calculate_shannon_entropy_scompression(adata)

        if adata.X.sum() == 0:
            logger.warn(f"The DGE from {dge_path} is empty")

        return adata


def load_external_dge(dge_path):
    import scanpy as sc

    from scanpy._utils import check_nonnegative_integers
    from scipy.sparse import issparse, csc_matrix
    from spacemake.errors import SpacemakeError

    adata = sc.read(dge_path)

    if not check_nonnegative_integers(adata.X):
        raise SpacemakeError(
            f"External dge seems to contain values "
            + "which are already normalised. Raw-count matrix expected."
        )

    if not issparse(adata.X):
        adata.X = csc_matrix(adata.X)

    # name the index
    adata.obs.index.name = "cell_bc"

    # attach metrics such as: total_counts, pct_mt_counts, etc
    # also attach n_genes, and calculate pcr
    calculate_adata_metrics(adata)

    return adata


def parse_barcode_file(barcode_file):
    import pandas as pd

    bc = pd.read_csv(barcode_file, sep="[,|\t]", engine='python')

    # rename columns
    bc = (
        bc.rename(
            columns={
                "xcoord": "x_pos",
                "ycoord": "y_pos",
                "barcodes": "cell_bc",
                "barcode": "cell_bc",
            }
        )
        .set_index("cell_bc")
        .loc[:, ["x_pos", "y_pos"]]
    )

    bc = bc.loc[~bc.index.duplicated(keep="first")]

    bc = bc.loc[~bc.index.duplicated(keep="first")]

    return bc


def attach_barcode_file(adata, barcode_file):
    bc = parse_barcode_file(barcode_file)

    # new obs has only the indices of the exact barcode matches
    new_obs = adata.obs.merge(bc, left_index=True, right_index=True, how="inner")
    adata = adata[new_obs.index, :]
    adata.obs = new_obs
    adata.obsm["spatial"] = adata.obs[["x_pos", "y_pos"]].to_numpy()

    return adata


def attach_puck_variables(adata, puck_variables):
    if "spatial" not in adata.obsm.keys():
        raise SpacemakeError(
            f"this dataset has no spatial information "
            + "available. Please attach the spatial information using the "
            + "spacemake.preprocess.attach_barcode_file() function first"
        )

    adata.uns["puck_variables"] = puck_variables

    x_pos_max, y_pos_max = tuple(adata.obsm["spatial"].max(axis=0))
    x_pos_min, y_pos_min = tuple(adata.obsm["spatial"].min(axis=0))
    #print(f"PUCK VARS {puck_variables} X MIN {x_pos_min} X MAX {x_pos_max} Y MIN {y_pos_min} Y MAX {y_pos_max}")

    width_um = adata.uns["puck_variables"]["width_um"]
    coord_by_um = (x_pos_max - x_pos_min) / width_um

    # this can be NaN if only one coordinate (only one cell, will fail)
    if coord_by_um > 0:
        height_um = int((y_pos_max - y_pos_min) / coord_by_um)
    else:
        height_um = 1 # avoid division by zero and error in reports
        coord_by_um = 1

    adata.uns["puck_variables"]["height_um"] = height_um
    adata.uns["puck_variables"]["coord_by_um"] = coord_by_um

    return adata


def attach_puck(adata, puck):
    attach_puck_variables(adata, puck.variables)
    adata.uns["puck_name"] = puck.name

    return adata


import os
from collections import defaultdict
from spacemake.errors import *
from spacemake.util import dotdict, wc_fill
import spacemake.snakemake.variables as smv

map_data = {
    # used to fetch all info needed to create a BAM file
    "MAP_RULES_LKUP": {},
    #   key: path of output BAM file
    #   value: dotdict with plenty of attributes
    # these support lookup by (project_id, sample_id)
    "ANNOTATED_BAMS": defaultdict(set),
    "REF_NAMES": defaultdict(set),
    "ALL_BAMS": defaultdict(set),
    "ALL_UNMAPPED_BAMS": defaultdict(set),
    "BAM_IS_NOT_TEMP": set(),
    "SAMPLE_MAP_STRATEGY": {},
    # used for symlink name to source mapping
    "BAM_SYMLINKS": {},
    "BAM_UNMAPPED_KEEP": set(),
    # used for automated mapping index generation
    "INDEX_FASTA_LKUP": {},
    # keep track of all used STAR indices to ensure they are properly unloaded after the run ended
    "STAR_INDICES": {},
    "REF_FOR_FINAL": {},
    #   key: bt2_index_file or star_index_file
    #   value: dotdict with
    #       .ref_path: path to genome FASTA
    #       .ann_path: path to annotation file (GTF) [optional]
    # needed for later stages of SPACEMAKE which require one "star.Log.final.out" file.
    "STAR_FINAL_LOG_SYMLINKS": {},
    # We create a symlink from the mapping that was accompanied by the 'final' identifier
}

#####################################
#### snakemake string templates #####
#####################################

ubam_input = "unaligned_bc_tagged.polyA_adapter_trimmed"
# this must be local and not have .bam appended!
# basically, if ubam_input were used as {target} in linked_bam it should eval to
# "unaligned_bc_tagged.polyA_adapter_trimmed.bam"

# The end-point. Usually annotation-tagged genome alignments
final_target = "final.polyA_adapter_trimmed"


#########################
### Default settings  ###
#########################

default_BT2_MAP_FLAGS = (
    " --local"
    " -L 10 -D 30 -R 30"
    " --ignore-quals"
    " --score-min=L,0,1.5"  # require 75% of perfect match (2=base match)
    " -k 10"
)
# original rRNA mapping code used --very-fast-local and that was that.

default_STAR_MAP_FLAGS = (
    # " --genomeLoad NoSharedMemory"
    " --outSAMprimaryFlag AllBestScore"
    " --outSAMattributes Standard"
    " --outSAMunmapped Within"
    " --outStd BAM_Unsorted"
    " --outSAMtype BAM Unsorted"
    " --limitOutSJcollapsed 5000000"
)

default_counting_flavor_with_annotation = "default"
default_counting_flavor_no_annotation = "custom_index"

#############################################################
##### Utility functions specific for the mapping module #####
#############################################################


def maybe_temporary(bam):
    return bam
    # DOES not currently work, as decision is made at parse time, before
    # any map-rules have been evaluated. Needs more thought.

    # # print("these are the NOT TEMP bam files", sorted(BAM_IS_NOT_TEMP))
    # if bam in map_data['BAM_IS_NOT_TEMP']:
    #     # print(bam, "is not temp!")
    #     return bam
    # else:
    #     # print(bam, "is TEMP!!!1!!11einself")
    #     return temp(bam)


def get_annotated_bams(wc):
    # print(wc)
    return {
        "annotated_bams": sorted(
            map_data["ANNOTATED_BAMS"][(wc.project_id, wc.sample_id)]
        )
    }


def get_all_mapped_bams(wc):
    # print(wc)
    return {"mapped_bams": sorted(map_data["ALL_BAMS"][(wc.project_id, wc.sample_id)])}


def get_count_flavor_str(wc):
    cflavors = []
    for bam in get_all_mapped_bams(wc)["mapped_bams"]:
        mr = map_data["MAP_RULES_LKUP"][bam]
        cflavors.append(f"{mr.ref_name}@{mr.cflavor}")

    return ",".join(cflavors)


##########################
### Core functionality ###
##########################


def validate_mapstr(mapstr, config={}, species=None):
    # print(f"validate_mapstr(mapstr={mapstr}, config={config}, species={species})")
    if species and config:
        species_d = config.get_variable("species", name=species)
    else:
        species_d = {}

    def process(token):
        """
        based on the left-hand side name (left) and the rule encoded in token,
        create one mapping rule and up to one symlink rule.
        """

        parts = token.split(":")
        if len(parts) == 2:
            mapper, ref = parts
            link_name = None
        elif len(parts) == 3:
            mapper, ref, link_name = parts
        else:
            raise ConfigVariableError(
                f"map_strategy contains a map-rule with unexpected number of parameters: {parts}"
            )

        # old format! We need to swap the order
        if ref.split("@")[0] in ["bowtie2", "STAR"]:
            ref, mapper = mapper, ref

        if species_d:
            if not ref in species_d:
                raise ConfigVariableNotFoundError(
                    ref,
                    f"reference name '{ref}' is not among the refs registered for {species}: {sorted(species_d.keys())}. Entire map-rule: {mapstr}",
                )

        if "@" in mapper:
            # we have a counting-flavor directive
            mapper, cflavor = mapper.split("@")
            if config:
                config.assert_variable("quant", cflavor)
        else:
            cflavor = None

        if cflavor:
            cfstr = f"@{cflavor}"
        else:
            cfstr = ""
        if link_name:
            lnkstr = f":{link_name}"
        else:
            lnkstr = ""

        return f"{mapper}{cfstr}:{ref}{lnkstr}"

    chain_rebuild = []
    for tokens in mapstr.split("->"):
        rebuilds = [process(r) for r in tokens.split(",")]
        chain_rebuild.append(",".join(rebuilds))

    return "->".join(chain_rebuild)


def mapstr_to_targets(mapstr, left="uBAM", final="final"):
    """
    Converts a mapping strategy provided as a string into a series of map rules, which translate into
    BAM names and their dependencies. Downstream, rule matching is guided by the convention of the
    strategy-defined BAM filenames ending in "STAR.bam" or "bowtie2.bam".

    Examples:

        rRNA:bowtie2->genome:STAR:final

    The input to the first mappings is always the uBAM - the CB and UMI tagged, pre-processed,
    but unmapped reads.
    map-rules are composed of two or three paramters, separated by ':'.
    The first two parameters for the mapping are <mapper>:<reference>. The target BAM will have
    the name <reference>.<mapper>.bam.
    Optionally, a triplet can be used <mapper>:<reference>:<symlink> where the presence of <symlink>
    indicates that the BAM file should additionally be made accessible under the name <symlink>.bam, a useful
    shorthand or common hook expected by other stages of SPACEMAKE. A special symlink name is "final"
    which is required by downstream stages of SPACEMAKE. If no "final" is specified, the last map-rule
    automatically is selected and symlinked as "final".
    Note, due to the special importance of "final", it may get modified to contain other flags of the run-mode
    that are presently essential for SPACEMAKE to have in the file name ("final.bam" may in fact be
    "final.polyA_adapter_trimmed.bam")

    The example above is going to create

        (1) rRNA.bowtie2.bam

            using bowtie2 and the index associated with the "rRNA" reference


        (2) genome.STAR.bam

            using STAR on the *unmapped* reads from BAM (1)


        (3) final.bam

            a symlink pointing to the actual BAM created in (2).

    Note that one BAM must be designated 'final.bam', or the last BAM file created will be selected as final.
    (used as input to downstream processing rules for DGE creation, etc.)

    NOTE: Parallel mappings can be implemented by using commata:

        bowtie2:rRNA,STAR:genome:final

        This rule differs from the first example because it will align the unmapped reads from the uBAM
        in parallel to the rRNA reference and to the genome. In this way the same reads can match to both
        indices.

    NOTE: Gene tagging will be applied automatically if annotation data were provided for the associated
    reference index (by using spacemake config add_species --annotation=... )
    """
    mapstr = validate_mapstr(mapstr)

    def process(token, left):
        """
        based on the left-hand side name (left) and the rule encoded in token,
        create one mapping rule and up to one symlink rule.
        """
        parts = token.split(":")
        mr = dotdict()
        lr = None

        if len(parts) == 2:
            mapper, ref = parts
            link_name = None
        elif len(parts) == 3:
            mapper, ref, link_name = parts
            link_name = link_name.replace("final", final)
        else:
            raise ValueError(
                f"map_strategy contains a map-rule with unexpected number of parameters: {parts}"
            )

        if "@" in mapper:
            # we have a counting-flavor directive
            mapper, cflavor = mapper.split("@")
        else:
            cflavor = "auto"

        mr.input_name = left
        mr.mapper = mapper
        mr.cflavor = cflavor
        mr.ref_name = ref
        mr.out_name = f"{ref}.{mapper}"
        mr.keep_unmapped = False

        if link_name:
            lr = dotdict(
                link_src=mr.out_name, link_name=link_name, ref_name=mr.ref_name
            )

        return mr, lr

    map_rules = []
    link_rules = []

    chain = mapstr.split("->")
    final_link = None

    while chain:
        right = chain.pop(0)
        if left == right:
            continue

        last_mr = []
        for r in right.split(","):
            mr, lr = process(r, left=left)
            map_rules.append(mr)
            last_mr.append(mr)

            if lr:
                link_rules.append(lr)
                # check if we have a "final" mapping
                if final in lr.link_name:
                    final_link = lr

        left = f"not_{mr.out_name}"
        # left = mr.out_name

    for mr in last_mr:
        mr.keep_unmapped = True

    if not final_link:
        # we need to manufacture a "final" link_rule by taking the last mapping
        last = map_rules[-1]
        link_rules.append(
            dotdict(link_src=last.out_name, link_name=final, ref_name=last.ref_name)
        )

    # print("generated the following map_rules")
    # for m in map_rules:
    #     print(m)
    return map_rules, link_rules


def get_mapped_BAM_output(
    project_df=None, config=None, default_strategy="genome:STAR:final"
):
    """
    This function is called from main.smk at least once
    to determine which output BAM files need to be generated and
    to parse the map_strategy into rules and dependencies.
    """
    out_files = []

    for index, row in project_df.df.iterrows():
        # rules so far are "local" to each sample. Here we create full paths, merging with
        # project_id/sample_id from the project_df

        map_strategy = getattr(row, "map_strategy", default_strategy)
        map_data["SAMPLE_MAP_STRATEGY"][index] = map_strategy

        map_rules, link_rules = mapstr_to_targets(
            map_strategy, left=ubam_input, final=final_target
        )

        species_d = project_df.config.get_variable("species", name=row.species)
        is_merged = project_df.get_metadata(
            "is_merged", project_id=index[0], sample_id=index[1]
        )
        if is_merged:
            continue

        for mr in map_rules:
            mr.project_id = index[0]
            mr.sample_id = index[1]
            mr.species = row.species

            out_files.append(wc_fill(smv.unmapped_removed_flag, mr))

            mr.out_path = wc_fill(smv.mapped_bam, mr)
            mr.out_unmapped_path = wc_fill(smv.unmapped_bam, mr)
            if mr.keep_unmapped:
                map_data["BAM_IS_NOT_TEMP"].add(mr.out_unmapped_path)

            mr.link_name = mr.input_name
            mr.input_path = wc_fill(smv.linked_bam, mr)

            mr.ref_path = species_d[mr.ref_name]["sequence"]
            mr.ann_path = species_d[mr.ref_name].get("annotation", None)
            map_data["ALL_BAMS"][(index[0], index[1])].add(mr.out_path)
            map_data["ALL_UNMAPPED_BAMS"][(index[0], index[1])].add(
                mr.out_unmapped_path
            )

            if mr.cflavor == "auto":
                # if we have annotation use the actual default, which works for complex
                # gene models
                # if we do NOT have annotation, assume custom_index (rRNA, spikes, CITE-seq tags etc.)
                mr.cflavor = (
                    default_counting_flavor_with_annotation
                    if mr.ann_path
                    else default_counting_flavor_no_annotation
                )

            if mr.ann_path:
                mr.ann_final = wc_fill(smv.species_reference_annotation, mr)
                mr.ann_final_compiled = wc_fill(
                    smv.species_reference_annotation_compiled, mr
                )
                mr.ann_final_compiled_target = wc_fill(
                    smv.species_reference_annotation_compiled_target, mr
                )

                # keep track of all annotated BAM files we are going to create
                # for subsequent counting into DGE matrices/h5ad
                map_data["ANNOTATED_BAMS"][(index[0], index[1])].add(mr.out_path)
                map_data["REF_NAMES"][(index[0], index[1])].add(mr.ref_name)
            else:
                mr.ann_final = []

            default_STAR_INDEX = wc_fill(smv.star_index, mr)
            default_BT2_INDEX = wc_fill(smv.bt2_index_param, mr)
            if mr.mapper == "bowtie2":
                mr.map_index_param = species_d[mr.ref_name].get(
                    "BT2_index", default_BT2_INDEX
                )  # the parameter passed on to the mapper
                mr.map_index = os.path.dirname(mr.map_index_param)  # the index_dir
                mr.map_index_file = (
                    mr.map_index_param + ".1.bt2"
                )  # file present if the index is actually there
                mr.map_flags = species_d[mr.ref_name].get(
                    "BT2_flags", default_BT2_MAP_FLAGS
                )

            elif mr.mapper == "STAR":
                mr.map_index = species_d[mr.ref_name].get(
                    "index_dir", default_STAR_INDEX
                )
                mr.map_index_param = mr.map_index
                mr.map_index_file = mr.map_index + "/SAindex"
                mr.star_idx_service = smv.star_idx_service.format(**mr)
                mr.map_flags = species_d[mr.ref_name].get(
                    "STAR_flags", default_STAR_MAP_FLAGS
                )

                map_data["STAR_INDICES"][(mr.species, mr.ref_name)] = mr.map_index_param

            map_data["MAP_RULES_LKUP"][mr.out_path] = mr
            map_data["INDEX_FASTA_LKUP"][mr.map_index_file] = mr
            # out_files.append(mr.out_path)

        # by convention we keep the last unmapped BAM file in a chain,
        # the others are temporary
        map_data["BAM_UNMAPPED_KEEP"].add(mr.out_unmapped_path)

        # process all symlink rules
        for lr in link_rules:
            lr.link_path = smv.linked_bam.format(
                project_id=index[0], sample_id=index[1], link_name=lr.link_name
            )
            lr.src_path = smv.linked_bam.format(
                project_id=index[0], sample_id=index[1], link_name=lr.link_src
            )
            map_data["BAM_SYMLINKS"][lr.link_path] = lr.src_path

            if lr.link_name == final_target:
                final_log_name = smv.star_log_file.format(
                    project_id=index[0], sample_id=index[1]
                )
                final_log = smv.star_target_log_file.format(
                    ref_name=lr.ref_name, project_id=index[0], sample_id=index[1]
                )
                # print("STAR_FINAL_LOG_SYMLINKS preparation", final_target, final_log_name, "->", final_log)
                map_data["STAR_FINAL_LOG_SYMLINKS"][final_log_name] = final_log
                map_data["REF_FOR_FINAL"][lr.link_path] = mr.ref_path

                out_files.append(lr.link_path)

    # for k, v in sorted(map_data["MAP_RULES_LKUP"].items()):
    #     print(f"map_rules for '{k}'")
    #     print(v)

    # print("BAM_SYMLINKS")
    # for k, v in map_data["BAM_SYMLINKS"].items():
    #     print(f"    output={k} <- source={v}")

    # print("out_files", out_files)
    return out_files



#!/usr/bin/env python
import unittest
import os
import os.path
import sys
import subprocess
import yaml
import pandas as pd

dropseq_tools = "/data/rajewsky/shared_bins/Drop-seq_tools-2.4.0"
base_dir = os.path.join(os.path.dirname(__file__), "..")

# set to true to track code coverage of the tests here using 'coverage'
code_coverage = False

test_species_data = [
    (
        "test_hsa",
        "genome",
        f"{base_dir}/test_data/test_genome.fa.gz",
        f"{base_dir}/test_data/test_genome.gtf.gz",
    ),
    (
        "test_hsa",
        "rRNA",
        f"{base_dir}/test_data/rRNA_hsa.fa.gz",
        f"",
    ),
    (
        "test_hsa",
        "miRNA",
        f"{base_dir}/test_data/mirgenedb.hsa.mature.fa.gz",
        f"{base_dir}/test_data/mirgenedb.hsa.mature.gtf.gz",
    ),
]
test_project_data = [
    (
        "test_hsa",
        "test",
        "test_01",
        f"{base_dir}/test_data/test_reads.R1.fastq.gz",
        f"{base_dir}/test_data/test_reads.R2.fastq.gz",
        "--map_strategy='STAR:genome:final'",
    ),
    (
        "test_hsa",
        "test",
        "test_02",
        f"{base_dir}/test_data/test_reads.R1.fastq.gz",
        f"{base_dir}/test_data/test_reads.R2.fastq.gz",
        "--map_strategy='bowtie2:rRNA->bowtie2:miRNA->STAR:genome:final'",
    ),
    (
        "test_hsa",
        "test",
        "test_03_nofinal",
        f"{base_dir}/test_data/test_reads.R1.fastq.gz",
        f"{base_dir}/test_data/test_reads.R2.fastq.gz",
        "--map_strategy='bowtie2:rRNA->bowtie2:miRNA->STAR:genome'",
    ),
]


def which_spacemake():
    p = subprocess.run("which spacemake", shell=True, capture_output=True)
    return p.stdout.decode("ascii").rstrip()


spacemake_cmd = which_spacemake()

if code_coverage:
    # gathering coverage information
    spacemake_cmd = f"coverage run -p {spacemake_cmd}"


def shell_bam_to_md5(path):
    "prints BAM content as SAM, sorts lexicographically and computes md5 hash"
    p = subprocess.run(
        f"samtools view {path} | sort | md5sum", shell=True, capture_output=True
    )
    md5 = p.stdout.split()[0]
    return (path, md5.decode("ascii"))


def gather_bam_hashes(path):
    results = {}
    for root, dirs, files in os.walk(path):
        for fname in files:
            if fname.endswith("bam"):
                bpath, md5 = shell_bam_to_md5(os.path.join(root, fname))
                results[bpath] = md5

    return results


def print_bam_hashes(results):
    for bpath, md5 in sorted(results.items()):
        print(f"{bpath}\t{md5}")


def load_bam_hashes(path):
    results = {}
    with open(path) as hfile:
        for line in hfile:
            bpath, md5 = line.rstrip().split("\t")
            results[bpath] = md5

    return results


class SpaceMakeCmdlineTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        os.system("mkdir -p _tests")
        os.chdir("_tests")

    @classmethod
    def tearDownClass(cls):
        # os.chdir("..")
        # os.system("rm -rf _tests")
        pass

    def run_spacemake(
        self,
        *args,
        expect_fail=False,
        check_returncode=True,
        check_success=True,
        check_stderr=True,
        **kwargs,
    ):
        print(
            f"running spacemake: cwd='{os.getcwd()}' args={args} kw={kwargs}", end=" "
        )
        p = subprocess.run(*args, shell=True, capture_output=True, **kwargs)
        print(f"returncode={p.returncode}")
        # print(f"\nrun_spacemake() results:\nstdout={p.stdout}\nstderr={p.stderr}")
        with open("run_spacemake.out.log", "ab") as out:
            if expect_fail:
                self.assertFalse(p.stdout.endswith(b"SUCCESS!\n"))
            else:
                out.write(bytes(f"RUN {args} \n", "ascii"))
                out.write(b"\nSTDOUT\n")
                out.write(p.stdout)
                out.write(b"\nSTDERR\n")
                out.write(p.stderr)
                out.write(bytes(f"RETURNCODE {p.returncode}\n", "ascii"))
                if check_returncode:
                    self.assertEqual(p.returncode, 0)

                if check_success:
                    self.assertTrue(p.stdout.endswith(b"SUCCESS!\n"))

                if check_stderr:
                    self.assertEqual(len(p.stderr), 0)

        return p

    def load_config(self):
        with open("config.yaml") as yamlfile:
            y = yaml.safe_load(yamlfile.read())

        return y

    def load_project_df(self):
        return pd.read_csv("project_df.csv")

    def add_species(self, name, ref, seq, ann, **kw):
        p = self.run_spacemake(
            f"{spacemake_cmd} config add_species"
            f" --name={name}"
            f" --reference={ref}"
            f" --sequence={seq}"
            f" --annotation={ann}",
            **kw,
        )
        return self.load_config()

    def add_genome_old(self, name, seq, ann, **kw):
        p = self.run_spacemake(
            f"{spacemake_cmd} config add_species"
            f" --name={name}"
            f" --genome={seq}"
            f" --annotation={ann}",
            **kw,
        )
        return self.load_config()

    def add_sample(self, species, pid, sid, r1, r2, options, **kw):
        p = self.run_spacemake(
            f"{spacemake_cmd} projects add_sample"
            f" --species={species}"
            f" --project_id={pid}"
            f" --sample_id={sid}"
            f" --R1={r1}"
            f" --R2={r2} {options}",
            check_stderr=False,
            **kw,
        )
        return self.load_project_df()

    def test_0_init(self):
        self.run_spacemake(f"{spacemake_cmd} init --dropseq_tools={dropseq_tools}", check_stderr=False)
        self.assertTrue(os.access("config.yaml", os.R_OK))

    def test_1_add_species(self):
        for name, ref, seq, ann in test_species_data:
            if name == "genome":
                # test backward-compatible --genome option
                y = self.add_genome_old(name, seq, ann, check_stderr=False)
            else:
                y = self.add_species(name, ref, seq, ann, check_stderr=False)

            self.assertTrue("species" in y)
            self.assertTrue(name in y["species"])
            self.assertTrue(ref in y["species"][name])
            self.assertEqual(y["species"][name][ref]["sequence"], seq)
            self.assertEqual(y["species"][name][ref]["annotation"], ann)

            # expect failure if trying to add same species, reference combination again
            y2 = self.add_species(name, ref, seq, ann, expect_fail=True, check_stderr=False)

            # expect unchanged config.yaml
            y1_str = yaml.dump(y)
            y2_str = yaml.dump(y2)
            self.assertEqual(y1_str, y2_str)

    def test_2_add_project(self):
        for species, pid, sid, r1, r2, options in test_project_data:
            df = self.add_sample(species, pid, sid, r1, r2, options)
            self.assertTrue(os.access("project_df.csv", os.R_OK))
            x = df.set_index("sample_id").loc[sid]
            self.assertEqual(x.species, species)
            self.assertEqual(x.project_id, pid)
            self.assertEqual(x.R1, str([r1]))
            self.assertEqual(x.R2, str([r2]))

            # expect failure if trying to add the same sample again
            df2 = self.add_sample(species, pid, sid, r1, r2, options, expect_fail=True)
            # expect unchanged project_df
            self.assertTrue(df.equals(df2))

    def test_6_update_sample(self):
        self.run_spacemake(
            f"{spacemake_cmd} projects update_sample "
            "--project_id=test --sample_id=test_01 "
            "--map_strategy='bowtie2:rRNA->STAR:genome'",
            check_stderr=False
        )
        self.run_spacemake(
            f"{spacemake_cmd} projects update_sample "
            "--project_id=test --sample_id=test_01 "
            "--map_strategy='STAR:genome'",
            check_stderr=False
        )

    def test_99_run(self):
        self.run_spacemake(
            f"{spacemake_cmd} run --cores=8", check_returncode=False, check_stderr=False
        )

#    def test_4_bamcheck(self):
#        # test correct BAM content
#        expect = load_bam_hashes("../test_data/test_bam_md5.txt")
#        for bpath, md5 in sorted(gather_bam_hashes(".").items()):
#            if bpath in expect:
#                print(f"checking '{bpath}'")
#                self.assertEqual(md5, expect[bpath])
#            else:
#                print(f"missing reference checksum for '{bpath}'- skipping test")
#
#        # TODO: test correct DGE content


if __name__ == "__main__":
    ## run this line once, together with output redirect to create
    ## reference md5 hashes from a run you deem correct
    # print_bam_hashes(gather_bam_hashes("."))
    unittest.main(verbosity=2)



import numpy as np
import logging
import pandas as pd

# import cutadapt.align
from collections import defaultdict

__version__ = "0.9"
__author__ = ["Marvin Jens"]
__license__ = "GPL"


def parse_cmdline():
    import argparse

    parser = argparse.ArgumentParser(
        description="tag alignments in a BAM file using various rules (aRNA and miRNA index)"
    )
    parser.add_argument(
        "bam_in",
        help="bam input (default=stdin)",
        default="/dev/stdin",
        # nargs="+",
    )
    parser.add_argument(
        "--bam-out",
        help="bam output (default=stdout)",
        default="/dev/stdout",
    )
    parser.add_argument(
        "--bam-out-mode",
        help="bam output mode (default=b)",
        default="b",
    )
    parser.add_argument(
        "--rules",
        help="which rules to apply. Options are 'aRNA','miRNA'",
        default="aRNA",
        choices=["aRNA", "miRNA", "genome"],
    )
    parser.add_argument(
        "--lookup",
        help="tab-separated table with reference -> function, name mappings",
        default="",
    )
    parser.add_argument(
        "--primer-pool",
        help="tab-separated table with reference -> function, name mappings",
        default="",
    )

    parser.add_argument(
        "--skim",
        help="skim through the BAM by investigating only every <skim>-th record (default=1 off)",
        default=1,
        type=int,
    )
    # parser.add_argument(
    #     "--min-length",
    #     help="minimal allowed read-length left after trimming (default=18)",
    #     type=int,
    #     default=18,
    # )
    # parser.add_argument(
    #     "--min-qual",
    #     help="minimal quality score for quality trimming (default=30)",
    #     type=int,
    #     default=20,
    # )
    parser.add_argument(
        "--stats-out",
        help="write tab-separated table with tagging overview results here",
        default="",
    )
    return parser.parse_args()


def make_header(bam):
    import os
    import sys

    header = bam.header.to_dict()
    progname = os.path.basename(__file__)
    # if "PG" in header:
    # for pg in header['PG']:
    #     if pg["ID"] == progname:
    #         progname = progname + ".1"

    pg_list = header.get("PG", [])
    pg = {
        "ID": progname,
        "PN": progname,
        "CL": " ".join(sys.argv[1:]),
        "VN": __version__,
    }
    if len(pg_list):
        pg["PP"] = pg_list[-1]["ID"]

    header["PG"] = pg_list + [pg]
    return header


class AbundantRNATagger:
    def __init__(self, bam, lkup_table={}):
        self.logger = logging.getLogger("AbundantRNATagger")
        self.lkup_table = lkup_table
        self.bam = bam
        self.tid_lkup = {}
        self.umi = set()

        self.counter = defaultdict(int)
        self.tag_names_to_count = ["af", "an"]

    def cached_name_for_tid(self, tid):
        if not tid in self.tid_lkup:
            self.tid_lkup[tid] = self.bam.get_reference_name(tid)
        return self.tid_lkup[tid]

    def make_tags(self, aln):
        name = self.cached_name_for_tid(aln.tid)
        _tags = aln.get_tags()
        tags = dict(_tags)
        key = (aln.query_sequence, tags.get("CB", "NA"), tags.get("MI", "NA"))

        if key in self.umi:
            ax = "PCR"
        else:
            ax = "UMI"

        self.umi.add(key)
        af, an = self.lkup_table.get(name, ("NA", "NA"))
        new_tags = [("ax", ax), ("an", an), ("af", af)]
        tags.update(dict(new_tags))

        return name, _tags + new_tags, tags

    def count(self, tags):
        self.counter[("reads", "stats", "N_reads")] += 1
        for n in self.tag_names_to_count:
            self.counter[("reads", n, tags[n])] += 1

        if tags["ax"] == "UMI":
            self.counter[("UMIs", "stats", "N_UMIs")] += 1
            for n in self.tag_names_to_count:
                self.counter[("UMIs", n, tags[n])] += 1

    def tag_alignment(self, aln):
        name, _tags, tags = self.make_tags(aln)
        self.count(tags)
        aln.set_tags(_tags)
        return aln, tags

    def write_stats(self, fname):
        N_reads = self.counter[("reads", "stats", "N_reads")]
        N_UMIs = self.counter[("UMIs", "stats", "N_UMIs")]

        with open(fname, "wt") as f:
            f.write("scope\ttag\tname\tcount\tpercent\n")
            for (scope, tag, name), v in sorted(
                self.counter.items(), key=lambda x: (x[0][:2], -x[1])
            ):
                if scope == "reads":
                    total = N_reads
                else:
                    total = N_UMIs
                f.write(f"{scope}\t{tag}\t{name}\t{v}\t{100.0 * v/total:.2f}\n")


class miRNATagger(AbundantRNATagger):
    def __init__(self, bam, lkup_table, lkup_targeted):
        AbundantRNATagger.__init__(self, bam, lkup_table)

        self.lkup_targeted = lkup_targeted
        self.lkup_targeted_gene = {}
        for key, pools in lkup_targeted.items():
            gene = self.lkup_table[key]
            self.lkup_targeted_gene[gene] = pools

        self.tag_names_to_count.append("aa")

    def make_tags(self, aln):
        name, _tags, tags = AbundantRNATagger.make_tags(self, aln)

        an = tags["an"]
        # parse CIGAR to get 5'/3' clipping and longest contiguous match
        n_match = 0
        n_clip5 = 0
        n_clip3 = 0
        if aln.cigartuples:
            last_i = len(aln.cigartuples) - 1
            for i, (op, n) in enumerate(aln.cigartuples):
                if op == 0:
                    n_match = max(n, n_match)

                if (i == 0) and (op == 4):
                    n_clip5 = n

                if (i == last_i) and (op == 4):
                    n_clip3 = n

        # populate "artifact warning" tag
        aa = []
        if "TSO" in tags.get("A5", ""):
            aa.append("TSO")
            if name in self.lkup_targeted:
                # aa.append(f"POOL:{self.lkup_targeted[name]}")
                aa.append(f"POOL")

            elif an in self.lkup_targeted_gene:
                aa.append(f"pool")

        if tags["af"] == "junk":
            aa.append("junk")

        if not "polyA" in tags.get("A3", ""):
            aa.append("no_polyA")

        if n_clip5 != 2:
            aa.append("no_lig_NN")

        if n_match < 17:
            aa.append("short_match")

        if n_clip3 > 10:
            aa.append("too_much_clipped")

        if not aa:
            aa.append("pass")

        aa = ",".join(aa)

        new_tags = [
            ("aa", aa),
            ("c5", n_clip5),
            ("c3", n_clip3),
            ("cm", n_match),
        ]

        tags.update(dict(new_tags))
        return (name, _tags + new_tags, tags)


class GenomeTagger(AbundantRNATagger):
    def select_genes(self, strand, gn_list, gf_list, gs_list):
        prio = {
            # sense-match, exonic
            (True, True): 0,
            (True, False): 1,
            (False, True): 20,
            (False, False): 30,
        }
        results = []
        best = 10000
        for i, (gn, gf, gs) in enumerate(zip(gn_list, gf_list, gs_list)):
            sense_match = gs == strand
            exonic = (gf == "UTR") or (gf == "CODING")
            p = prio[(sense_match, exonic)]
            results.append((p, gn, i))
            best = min(best, p)

        genes = set()
        gf = ""
        for p, gn, i in results:
            if p == best:
                genes.add(gn)
                gf = gf_list[i]

        genes = list(genes)
        if len(genes) == 1:
            return genes[0], gf
        else:
            return ",".join(sorted(genes)), gf

    def make_tags(self, aln):
        # name = aln.get_tag("gn")
        _tags = aln.get_tags()
        tags = dict(_tags)

        name, af = self.select_genes(
            "-" if aln.is_reverse else "+",
            tags.get("gn", "").split(","),
            tags.get("gf", "").split(","),
            tags.get("gs", "").split(","),
        )

        key = (aln.query_sequence[:20], tags.get("CB", "NA"), tags.get("MI", "NA"))

        if key in self.umi:
            ax = "PCR"
        else:
            ax = "UMI"

        self.umi.add(key)

        new_tags = [("ax", ax), ("an", name), ("af", af)]
        tags.update(dict(new_tags))

        return name, _tags + new_tags, tags


if __name__ == "__main__":
    args = parse_cmdline()

    stats = defaultdict(int)
    total = defaultdict(int)
    lhist = defaultdict(int)

    import pysam

    bam_in = pysam.AlignmentFile(args.bam_in, "rb", check_sq=False)
    bam_out = pysam.AlignmentFile(
        args.bam_out, f"w{args.bam_out_mode}", header=make_header(bam_in)
    )

    lkup_table = {}
    if args.lookup:
        for row in pd.read_csv(args.lookup, sep="\t").itertuples():
            lkup_table[row.ref] = (row.af, row.an)

    lkup_pool = {}
    if args.primer_pool:
        for row in pd.read_csv(args.primer_pool, sep="\t").itertuples():
            txt = []
            # print(row)
            if row._2:
                txt.append("3")
            if row._3:
                txt.append("12")
            if row._4:
                txt.append("22")
            if row._5:
                txt.append("mmu")

            lkup_pool[row.name] = ",".join(txt)

    if args.rules == "aRNA":
        tagger = AbundantRNATagger(bam_in, lkup_table)
    elif args.rules == "miRNA":
        tagger = miRNATagger(bam_in, lkup_table, lkup_pool)
    else:
        tagger = GenomeTagger(bam_in)

    for i, aln in enumerate(bam_in.fetch(until_eof=True)):
        if args.skim and i % args.skim != 0:
            continue

        aln, tags = tagger.tag_alignment(aln)
        bam_out.write(aln)

    if args.stats_out:
        tagger.write_stats(args.stats_out)



################################
# Final output file generation #
################################
def get_output_files(
    pattern,
    projects=[],
    samples=[],
    filter_merged=False,
    run_on_external=True,
    puck_barcode_file_matching_type="none",
    check_puck_collection=False,
    qc=False,
    **kwargs,
):
    out_files = []
    df = project_df.df

    if projects != [] or samples != []:
        project_df.assert_index_value(projects, "project_id")
        project_df.assert_index_value(samples, "sample_id")

        ix = project_df.get_ix_from_project_sample_list(
            project_id_list=projects, sample_id_list=samples
        )

        df = df.loc[ix]

    if filter_merged:
        df = df.loc[~df.is_merged]

    for index, row in df.iterrows():
        project_id, sample_id = index

        is_external = project_df.is_external(project_id=project_id, sample_id=sample_id)

        has_dge = project_df.has_dge(project_id=project_id, sample_id=sample_id)

        if not has_dge:
            # if there is no dge, skip
            continue

        if not run_on_external and is_external:
            # if we do not want to run this on external samples
            # and sample is external, skip
            continue

        if puck_barcode_file_matching_type == "none":
            # reset to empty string
            puck_barcode_file_ids = []
        elif puck_barcode_file_matching_type == "spatial":
            # get only spatial
            puck_barcode_file_ids = project_df.get_puck_barcode_ids_and_files(
                project_id, sample_id
            )[0]
        elif puck_barcode_file_matching_type == "spatial_matching":
            puck_barcode_file_ids = project_df.get_matching_puck_barcode_file_ids(
                project_id=project_id, sample_id=sample_id
            )

        if check_puck_collection:
            puck_vars = project_df.get_puck_variables(
                project_id=project_id, sample_id=sample_id
            )

            if (len(puck_barcode_file_ids) == 0) or (
                "coordinate_system" not in puck_vars.keys()
            ):
                continue

            if (len(puck_barcode_file_ids) > 0) and (
                puck_barcode_file_ids[0] == "no_spatial_data"
            ):
                continue

            coordinate_system = puck_vars["coordinate_system"]
            if coordinate_system == "":
                continue

            puck_barcode_file_ids = "puck_collection"

        else:
            # add the non spatial barcode by default
            non_spatial_pbf_id = project_df.project_df_default_values[
                "puck_barcode_file_id"
            ][0]

            if non_spatial_pbf_id not in puck_barcode_file_ids:
                puck_barcode_file_ids.append(non_spatial_pbf_id)

        for run_mode in row["run_mode"]:
            run_mode_variables = project_df.config.get_run_mode(run_mode).variables
            if "polyA_adapter_trimmed" in kwargs:
                polyA_adapter_trimmed = kwargs["polyA_adapter_trimmed"]
            else:
                if run_mode_variables["polyA_adapter_trimming"]:
                    polyA_adapter_trimmed = ".polyA_adapter_trimmed"
                else:
                    polyA_adapter_trimmed = ""

            out_files = out_files + expand(
                pattern,
                project_id=project_id,
                sample_id=sample_id,
                puck_barcode_file_id=puck_barcode_file_ids,
                puck_barcode_file_id_qc=puck_barcode_file_ids,
                run_mode=run_mode,
                umi_cutoff=run_mode_variables["umi_cutoff"],
                polyA_adapter_trimmed=polyA_adapter_trimmed,
                **kwargs,
            )

    return out_files


def get_output_files_qc(
    pattern,
    projects=[],
    samples=[],
    filter_merged=False,
    run_on_external=True,
    puck_barcode_file_matching_type="none",
    **kwargs,
):
    _out_files = []

    # get files with puck_barcode_file_id (individual tiles)
    _out_files += get_output_files(
        pattern=pattern,
        projects=projects,
        samples=samples,
        filter_merged=filter_merged,
        run_on_external=run_on_external,
        puck_barcode_file_matching_type=puck_barcode_file_matching_type,
        check_puck_collection=False,
        **kwargs,
    )

    # get files for puck_collection
    _out_files += get_output_files(
        pattern=pattern,
        projects=projects,
        samples=samples,
        filter_merged=filter_merged,
        run_on_external=run_on_external,
        puck_barcode_file_matching_type=puck_barcode_file_matching_type,
        check_puck_collection=True,
        **kwargs,
    )

    return _out_files


def get_prealignment_files(pattern, filter_merged=False):
    df = project_df.df

    if filter_merged:
        df = df.loc[~df.is_merged]

    prealignment_files = []

    for index, _ in df.iterrows():
        project_id, sample_id = index

        run_mode_name = project_df.get_metadata(
            "run_mode", project_id=project_id, sample_id=sample_id
        )[0]

        run_mode_vars = project_df.config.get_run_mode(run_mode_name).variables
        puck_variables = project_df.get_puck_variables(
            project_id=index[0], sample_id=index[1]
        )

        if (
            run_mode_vars["spatial_barcode_min_matches"] == 0
            or puck_variables["coordinate_system"] == ""
        ):
            continue

        # TODO: does this need to be separated per run mode?
        prealignment_files.append(
            expand(
                pattern,
                project_id=project_id,
                sample_id=sample_id,
            )
        )

    return prealignment_files


def get_all_dges(wildcards):
    df = project_df.df

    dges = []

    for index, row in df.iterrows():
        project_id, sample_id = index
        puck_barcode_file_ids = project_df.get_matching_puck_barcode_file_ids(
            project_id=project_id, sample_id=sample_id
        )

        for run_mode in row["run_mode"]:
            if project_df.has_dge(project_id=project_id, sample_id=sample_id):
                for pbf_id in puck_barcode_file_ids:
                    dges.append(
                        get_dge_from_run_mode(
                            project_id=project_id,
                            sample_id=sample_id,
                            run_mode=run_mode,
                            data_root_type="complete_data",
                            downsampling_percentage="",
                            puck_barcode_file_id=pbf_id,
                        )["dge"],
                    )

    return dges


def get_all_dges_collection(wildcards):
    import os

    df = project_df.df

    dges = []

    for index, row in df.iterrows():
        project_id, sample_id = index

        puck_vars = project_df.get_puck_variables(
            project_id=project_id, sample_id=sample_id
        )

        puck_barcode_file_ids = project_df.get_puck_barcode_ids_and_files(
            project_id, sample_id
        )[0]

        if (len(puck_barcode_file_ids) == 0) or (
            "coordinate_system" not in puck_vars.keys()
        ):
            continue

        coordinate_system = puck_vars["coordinate_system"]
        if coordinate_system == "":
            continue

        if not os.path.exists(coordinate_system):
            raise FileNotFoundError(
                f"at project {project_id} sample {sample_id} "
                + f"'coordinate_system' file {coordinate_system} cannot be found"
            )

        # will consider puck_collection within all dges if a coordinate system is specified in the puck
        with_puck_collection = False if not coordinate_system else True

        for run_mode in row["run_mode"]:
            if project_df.has_dge(project_id=project_id, sample_id=sample_id):
                if with_puck_collection:
                    dges.append(
                        get_dge_collection_from_run_mode(
                            project_id=project_id,
                            sample_id=sample_id,
                            run_mode=run_mode,
                            data_root_type="complete_data",
                            downsampling_percentage="",
                        )["dge"]
                    )

    return dges


def get_raw_dge(wildcards):
    is_external = project_df.is_external(
        project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )

    has_dge = project_df.has_dge(
        project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )

    if has_dge:
        out_files = {}
    else:
        return []

    if is_external:
        out_files["dge"] = project_df.get_metadata(
            "dge", project_id=wildcards.project_id, sample_id=wildcards.sample_id
        )
    else:
        out_files["dge"] = dge_out
        out_files["dge_summary"] = dge_out_summary

    return out_files


def get_reads(wildcards):
    ###
    # R1 and R2 for demultiplexed reads will return none
    ###
    reads = project_df.get_metadata(
        "R" + wildcards.mate,
        sample_id=wildcards.sample_id,
        project_id=wildcards.project_id,
    )
    if reads is None or reads == []:
        return ["none"]
    else:
        # reads already
        return reads


# barcode flavor parsing and query functions
class dotdict(dict):
    """dot.notation access to dictionary attributes"""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __str__(self):
        buf = ["dotdict"]
        for k, v in self.items():
            if not k.startswith("__"):
                buf.append(f"  {k} = {v}")

        return "\n".join(buf)


def parse_barcode_flavors(
    config,
    bc_default_settings=dict(
        bc1_ref="",
        bc2_ref="",
        cell_raw="None",
        score_threshold=0.0,
        min_opseq_score=22,
        bam_tags="CR:{cell},CB:{cell},MI:{UMI}",
    ),
):
    """
    Reads the 'barcode_flavor' from 'knowledge' section of the config.yaml.
    parses and gathers the settings for barcode flavors
    """
    preprocess_settings = {}
    for flavor, flavor_settings in config["barcode_flavors"].items():
        # for each flavor, also retrieve the configuration
        # first make a copy of the default values
        d = dict(bc_default_settings)
        d.update(flavor_settings)
        preprocess_settings[flavor] = dotdict(d)

    res = dotdict(
        dict(
            preprocess_settings=preprocess_settings,
        )
    )

    return res


# all barcode flavor info from config.yaml
# is kept here for convenient lookup
bc_flavor_data = parse_barcode_flavors(config)


def get_bc_preprocess_settings(wildcards):
    """
    This function will return a dictionary of information
    on the read1 preprocessing, according to barcode_flavor
    """
    flavor = project_df.get_metadata(
        "barcode_flavor", project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )
    if flavor not in bc_flavor_data.preprocess_settings:
        raise Exception(flavor)

    settings = bc_flavor_data.preprocess_settings[flavor]

    adapter_flavor = project_df.get_metadata(
        "adapter_flavor", project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )

    settings["adapter_flavor"] = adapter_flavor
    return settings


def get_demux_indicator(wildcards):
    demux_dir = project_df.get_metadata(
        "demux_dir", sample_id=wildcards.sample_id, project_id=wildcards.project_id
    )

    return expand(demux_indicator, demux_dir=demux_dir)


def get_star_input_bam(wildcards):
    if wildcards.polyA_adapter_trimmed == ".polyA_adapter_trimmed":
        return {"reads": tagged_polyA_adapter_trimmed_bam}
    else:
        return {"reads": tagged_bam}


def is_merged(wildcards):
    return project_df.get_metadata(
        "is_merged", project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )


def get_final_bam(wildcards):
    if is_merged(wildcards):
        res = [final_merged_bam]
    else:
        res = [final_bam]

    return res


def get_final_bam_reference(wildcards):
    from spacemake.map_strategy import map_data

    if is_merged(wildcards):
        merged_from = project_df.get_metadata(
            "merged_from",
            project_id=wildcards.project_id,
            sample_id=wildcards.sample_id,
        )
        # collect the main reference to which the constituent
        # samples were aligned/annotated and make sure it is
        # the same.
        refs = set()
        wc = dotdict(wildcards)
        for project_id, sample_id in merged_from:
            wc.project_id = project_id
            wc.sample_id = sample_id
            fb = wc_fill(final_bam, wc)
            ref = map_data["REF_FOR_FINAL"][fb]
            # print(f">>> merged_from: {wc.sample_id} {fb} {ref}")
            refs.add(ref)

        assert len(refs) == 1
        ref = refs.pop()
    else:
        fb = wc_fill(final_bam, wildcards)
        ref = map_data["REF_FOR_FINAL"][fb]
        # print(f">>> {fb} {ref}")

    return ref


def get_dge_input_bam(wildcards):
    if wildcards.data_root_type == "complete_data":
        final_bam_pipe = final_bam_mm_included_pipe
        final_bam = get_final_bam(wildcards)
    elif wildcards.data_root_type == "downsampled_data":
        final_bam_pipe = downsampled_bam_mm_included_pipe
        final_bam = downsampled_bam

    if wildcards.mm_included == ".mm_included":
        out = {"reads": final_bam_pipe}
    else:
        out = {"reads": final_bam}

    return out


# TODO: rename to get_species_ref_sequence()
def get_species_genome(wildcards, ref="genome"):
    # This function will return the genome of a sample
    #    - annotation (.gtf file)
    #    - genome (.fa file)
    if "species" not in wildcards.keys():
        species = project_df.get_metadata(
            "species", project_id=wildcards.project_id, sample_id=wildcards.sample_id
        )
    else:
        species = wildcards.species

    files = project_df.config.get_variable("species", name=species)

    return [files[ref]["sequence"]]


def get_species_annotation(wildcards, ref="genome"):
    # This function will return the genome of a sample
    #    - annotation (.gtf file)
    #    - genome (.fa file)
    if "species" not in wildcards.keys():
        species = project_df.get_metadata(
            "species", project_id=wildcards.project_id, sample_id=wildcards.sample_id
        )
    else:
        species = wildcards.species

    files = project_df.config.get_variable("species", name=species)

    return [files[ref]["annotation"]]


def get_species_genome_annotation(wildcards, ref="genome"):
    # This function will return 2 things required by STAR:
    #    - annotation (.gtf file)
    #    - genome (.fa file)
    if "species" not in wildcards.keys():
        species = project_df.get_metadata(
            "species", project_id=wildcards.project_id, sample_id=wildcards.sample_id
        )
    else:
        species = wildcards.species

    return {
        "genome": species_genome.format(species=species),
        "annotation": species_annotation.format(species=species),
    }


def get_star_index(wildcards, ref="genome"):
    # This function will return 1 things required by STAR:
    #    - index directory
    species = project_df.get_metadata(
        "species", project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )
    species_data = project_df.config.get_variable("species", name=species)[ref]

    if "STAR_index_dir" in species_data:
        return {"index": species_data["STAR_index_dir"]}
    else:
        return {"index": expand(star_index, species=species)[0]}


# TODO: refactor into map_strategy and/or delete
def get_rRNA_genome(wildcards):
    files = project_df.config.get_variable("species", name=wildcards.species)

    return [files["rRNA_genome"]]


# TODO: refactor into map_strategy and/or delete
def get_bt2_rRNA_index(wildcards):
    species = project_df.get_metadata(
        "species", project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )

    files = project_df.config.get_variable("species", name=species)

    if "rRNA_genome" in files:
        return {"index": expand(bt2_rRNA_index_dir, species=species)[0]}

    return []


def get_run_modes_from_sample(project_id, sample_id):
    run_mode_names = project_df.get_metadata(
        "run_mode", project_id=project_id, sample_id=sample_id
    )

    run_modes = {}

    for run_mode in run_mode_names:
        run_modes[run_mode] = project_df.config.get_run_mode(run_mode).variables

    return run_modes


def get_dge_extra_params(wildcards):
    dge_type = wildcards.dge_type

    extra_params = ""

    if dge_type == ".exon":
        extra_params = ""
    elif dge_type == ".intron":
        extra_params = "LOCUS_FUNCTION_LIST=null LOCUS_FUNCTION_LIST=INTRONIC"
    elif dge_type == ".all":
        extra_params = "LOCUS_FUNCTION_LIST=INTRONIC"
    if dge_type == ".Reads_exon":
        extra_params = "OUTPUT_READS_INSTEAD=true"
    elif dge_type == ".Reads_intron":
        extra_params = (
            "OUTPUT_READS_INSTEAD=true LOCUS_FUNCTION_LIST=null"
            + "LOCUS_FUNCTION_LIST=INTRONIC"
        )
    elif dge_type == ".Reads_all":
        extra_params = "OUTPUT_READS_INSTEAD=true LOCUS_FUNCTION_LIST=INTRONIC"

    if wildcards.mm_included == ".mm_included":
        extra_params = extra_params + " READ_MQ=0"

    return extra_params


def get_files_to_merge(pattern, project_id, sample_id, **kwargs):
    # recursive function to find all files to merge. a merged sample can be merged
    # from merged samples. to avoid cyclic dependencies, here we look for all files
    # which are the dependencies of the underlying samples
    is_merged = project_df.get_metadata(
        "is_merged", project_id=project_id, sample_id=sample_id
    )

    files = []

    if not is_merged:
        files = expand(pattern, sample_id=sample_id, project_id=project_id, **kwargs)
    else:
        merge_ix = project_df.get_metadata(
            "merged_from", sample_id=sample_id, project_id=project_id
        )

        for p, s in merge_ix:
            files = files + get_files_to_merge(
                project_id=p, sample_id=s, pattern=pattern, **kwargs
            )

    return list(set(files))


def get_files_to_merge_snakemake(pattern):
    # inner function to be returned
    def get_merged_pattern(wildcards):
        kwargs = {}

        # konvert wildcards to dict
        for key, value in wildcards.items():
            kwargs[key] = value

        files = get_files_to_merge(pattern=pattern, **kwargs)

        return files

    return get_merged_pattern


def get_ribo_depletion_log(wildcards):
    is_merged = project_df.get_metadata(
        "is_merged", sample_id=wildcards.sample_id, project_id=wildcards.project_id
    )

    if is_merged:
        return [merged_ribo_depletion_log]
    else:
        return [ribo_depletion_log]


def get_top_barcodes(wildcards):
    if wildcards.n_beads == "spatial":
        return {"top_barcodes": spatial_barcodes}  # experimental
    if wildcards.dge_cleaned == "":
        return {"top_barcodes": top_barcodes}
    else:
        return {"top_barcodes": top_barcodes_clean}


def get_parsed_puck_file(wildcards):
    is_spatial = project_df.is_spatial(
        project_id=wildcards.project_id,
        sample_id=wildcards.sample_id,
        puck_barcode_file_id=wildcards.puck_barcode_file_id_qc,
    )

    if is_spatial or wildcards.puck_barcode_file_id_qc == "puck_collection":
        return parsed_spatial_barcodes.format(
            project_id=wildcards.project_id,
            sample_id=wildcards.sample_id,
            puck_barcode_file_id=wildcards.puck_barcode_file_id_qc,
        )
    else:
        return []


def get_dge_from_run_mode(
    project_id,
    sample_id,
    run_mode,
    data_root_type,
    downsampling_percentage,
    puck_barcode_file_id,
    only_spatial=False,
    to_mesh=None,
):
    has_dge = project_df.has_dge(project_id=project_id, sample_id=sample_id)

    if not has_dge:
        raise SpacemakeError(
            f"Sample with id (project_id, sample_id)={project_id}, {sample_id})"
            + f" does not have a DGE"
        )

    is_spatial = project_df.is_spatial(
        project_id=project_id,
        sample_id=sample_id,
        puck_barcode_file_id=puck_barcode_file_id,
    )

    is_external = project_df.is_external(project_id=project_id, sample_id=sample_id)

    run_mode_variables = project_df.config.get_run_mode(run_mode).variables

    dge_type = ""
    dge_cleaned = ""
    polyA_adapter_trimmed = ""
    mm_included = ""

    # assign wildcards only for internal samples
    if not is_external:
        if run_mode_variables["polyA_adapter_trimming"]:
            polyA_adapter_trimmed = ".polyA_adapter_trimmed"

        if run_mode_variables["count_intronic_reads"]:
            dge_type = ".all"
        else:
            dge_type = ".exon"

        if run_mode_variables["count_mm_reads"]:
            mm_included = ".mm_included"

        if run_mode_variables["clean_dge"]:
            dge_cleaned = ".cleaned"

    if run_mode_variables["mesh_type"] == "hexagon":
        spot_diameter_um = run_mode_variables["mesh_spot_diameter_um"]
        spot_distance_um = "hexagon"
    elif run_mode_variables["mesh_type"] == "circle":
        spot_diameter_um = run_mode_variables["mesh_spot_diameter_um"]
        spot_distance_um = run_mode_variables["mesh_spot_distance_um"]

    external_wildcard = ""

    n_beads = run_mode_variables["n_beads"]

    if is_external:
        n_beads = "external"
        external_wildcard = ".external"

    if is_spatial or only_spatial:
        n_beads = "spatial"

    # select which pattern
    # if sample is not spatial, we simply select the normal, umi_filtered
    # dge, with the top_n barcodes
    # otherwise, if sample is spatial, either we return he whole dge, containing
    # all beads, or the a meshgrid
    if not is_spatial and not only_spatial:
        dge_out_pattern = dge_out_h5ad
        dge_out_summary_pattern = dge_out_h5ad_obs
    elif run_mode_variables["mesh_data"] or to_mesh:
        dge_out_pattern = dge_spatial_mesh
        dge_out_summary_pattern = dge_spatial_mesh_obs
    elif to_mesh is not None and to_mesh == False:
        dge_out_pattern = dge_spatial
        dge_out_summary_pattern = dge_spatial_obs
    else:
        dge_out_pattern = dge_spatial
        dge_out_summary_pattern = dge_spatial_obs

    out_files_pattern = {"dge_summary": dge_out_summary_pattern, "dge": dge_out_pattern}

    out_files = {
        key: expand(
            pattern,
            project_id=project_id,
            sample_id=sample_id,
            dge_type=dge_type,
            dge_cleaned=dge_cleaned,
            polyA_adapter_trimmed=polyA_adapter_trimmed,
            mm_included=mm_included,
            spot_diameter_um=spot_diameter_um,
            spot_distance_um=spot_distance_um,
            n_beads=n_beads,
            is_external=external_wildcard,
            data_root_type=data_root_type,
            downsampling_percentage=downsampling_percentage,
            puck_barcode_file_id=puck_barcode_file_id,
        )
        for key, pattern in out_files_pattern.items()
    }

    return out_files


def get_dge_collection_from_run_mode(
    project_id,
    sample_id,
    run_mode,
    data_root_type,
    downsampling_percentage,
):
    has_dge = project_df.has_dge(project_id=project_id, sample_id=sample_id)

    if not has_dge:
        raise SpacemakeError(
            f"Sample with id (project_id, sample_id)={project_id}, {sample_id})"
            + f" does not have a DGE"
        )

    is_external = project_df.is_external(project_id=project_id, sample_id=sample_id)

    run_mode_variables = project_df.config.get_run_mode(run_mode).variables

    dge_type = ""
    dge_cleaned = ""
    polyA_adapter_trimmed = ""
    mm_included = ""

    # assign wildcards only for internal samples
    if not is_external:
        if run_mode_variables["polyA_adapter_trimming"]:
            polyA_adapter_trimmed = ".polyA_adapter_trimmed"

        if run_mode_variables["count_intronic_reads"]:
            dge_type = ".all"
        else:
            dge_type = ".exon"

        if run_mode_variables["count_mm_reads"]:
            mm_included = ".mm_included"

        if run_mode_variables["clean_dge"]:
            dge_cleaned = ".cleaned"

    if run_mode_variables["mesh_type"] == "hexagon":
        spot_diameter_um = run_mode_variables["mesh_spot_diameter_um"]
        spot_distance_um = "hexagon"
    elif run_mode_variables["mesh_type"] == "circle":
        spot_diameter_um = run_mode_variables["mesh_spot_diameter_um"]
        spot_distance_um = run_mode_variables["mesh_spot_distance_um"]

    external_wildcard = ""

    n_beads = run_mode_variables["n_beads"]

    if is_external:
        n_beads = "external"
        external_wildcard = ".external"

    n_beads = "spatial"

    if run_mode_variables["mesh_data"]:
        dge_out_pattern = dge_spatial_collection_mesh
        dge_out_summary_pattern = dge_spatial_collection_mesh_obs
    else:
        dge_out_pattern = dge_spatial_collection
        dge_out_summary_pattern = dge_spatial_collection_obs

    out_files_pattern = {"dge_summary": dge_out_summary_pattern, "dge": dge_out_pattern}

    out_files = {
        key: expand(
            pattern,
            project_id=project_id,
            sample_id=sample_id,
            dge_type=dge_type,
            dge_cleaned=dge_cleaned,
            polyA_adapter_trimmed=polyA_adapter_trimmed,
            mm_included=mm_included,
            spot_diameter_um=spot_diameter_um,
            spot_distance_um=spot_distance_um,
            n_beads=n_beads,
            is_external=external_wildcard,
            data_root_type=data_root_type,
            downsampling_percentage=downsampling_percentage,
        )
        for key, pattern in out_files_pattern.items()
    }

    return out_files


def get_all_barcode_readcounts(wildcards, prealigned=False):
    # returns all available barcode readcounts
    project_id = wildcards.project_id
    sample_id = wildcards.sample_id

    is_merged = project_df.get_metadata(
        "is_merged", project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )

    run_modes = get_run_modes_from_sample(wildcards.project_id, wildcards.sample_id)

    is_polyA_adapter_trimmed = set(
        [x["polyA_adapter_trimming"] for x in run_modes.values()]
    )

    # if sample has both polyA trimmed and untrimmed mapped bam files
    if len(is_polyA_adapter_trimmed) == 2:
        polyA_adapter_trimmed_wildcard = ["", ".polyA_adapter_trimmed"]
    elif True in is_polyA_adapter_trimmed:
        polyA_adapter_trimmed_wildcard = [".polyA_adapter_trimmed"]
    elif False in is_polyA_adapter_trimmed:
        polyA_adapter_trimmed_wildcard = [""]

    extra_args = {
        "sample_id": wildcards.sample_id,
        "project_id": wildcards.project_id,
        "polyA_adapter_trimmed": polyA_adapter_trimmed_wildcard,
    }

    if prealigned or is_merged:
        return {"bc_readcounts": expand(barcode_readcounts, **extra_args)}
    else:
        return {"bc_readcounts": expand(barcode_readcounts_prealigned, **extra_args)}


def get_qc_sheet_input_files(wildcards):
    # returns star_log, reads_type_out, strand_info
    # first checks the run modes, and returns either polyA_adapter_trimmed, untrimmed
    # or both
    project_id = wildcards.project_id
    sample_id = wildcards.sample_id

    is_merged = project_df.get_metadata(
        "is_merged", project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )

    run_modes = get_run_modes_from_sample(wildcards.project_id, wildcards.sample_id)

    is_polyA_adapter_trimmed = set(
        [x["polyA_adapter_trimming"] for x in run_modes.values()]
    )

    # if sample has both polyA trimmed and untrimmed mapped bam files
    if len(is_polyA_adapter_trimmed) == 2:
        polyA_adapter_trimmed_wildcard = ["", ".polyA_adapter_trimmed"]
    elif True in is_polyA_adapter_trimmed:
        polyA_adapter_trimmed_wildcard = [".polyA_adapter_trimmed"]
    elif False in is_polyA_adapter_trimmed:
        polyA_adapter_trimmed_wildcard = [""]

    extra_args = {
        "sample_id": wildcards.sample_id,
        "project_id": wildcards.project_id,
        "polyA_adapter_trimmed": polyA_adapter_trimmed_wildcard,
    }

    if is_merged:
        star_log_pattern = merged_star_log_file
    else:
        star_log_pattern = star_log_file

    to_return = {
        "star_log": expand(star_log_pattern, **extra_args),
        "reads_type_out": expand(reads_type_out, **extra_args),
        "strand_info": expand(strand_info, **extra_args),
    }

    for run_mode in run_modes:
        if wildcards.puck_barcode_file_id_qc != "puck_collection":
            run_mode_dge = get_dge_from_run_mode(
                project_id,
                sample_id,
                run_mode,
                data_root_type=wildcards.data_root_type,
                downsampling_percentage=wildcards.downsampling_percentage,
                puck_barcode_file_id=wildcards.puck_barcode_file_id_qc,
            )
        else:
            run_mode_dge = get_dge_collection_from_run_mode(
                project_id,
                sample_id,
                run_mode,
                data_root_type=wildcards.data_root_type,
                downsampling_percentage=wildcards.downsampling_percentage,
            )

        to_return[f"{run_mode}.dge_summary"] = run_mode_dge["dge_summary"]
        to_return[f"{run_mode}.dge"] = run_mode_dge["dge"]

    return to_return


def get_bam_tag_names(
    project_id, sample_id, default_tags="CR:{cell},CB:{cell},MI:{UMI},RG:{assigned}"
):
    barcode_flavor = project_df.get_metadata(
        "barcode_flavor", project_id=project_id, sample_id=sample_id
    )

    bam_tags = config["barcode_flavors"][barcode_flavor].get("bam_tags", default_tags)

    tag_names = {}

    for tag in bam_tags.split(","):
        tag_name, tag_variable = tag.split(":")

        tag_names[tag_variable] = tag_name

    return tag_names


def get_puck_file(wildcards):
    puck_barcode_file = project_df.get_puck_barcode_file(
        project_id=wildcards.project_id,
        sample_id=wildcards.sample_id,
        puck_barcode_file_id=wildcards.puck_barcode_file_id,
    )

    if puck_barcode_file is None:
        return []
    else:
        return {"barcode_file": puck_barcode_file}


def get_all_puck_files(wildcards):
    _, pbf = project_df.get_puck_barcode_ids_and_files(
        project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )

    return {"barcode_file": pbf}


def get_puck_collection_stitching_input(wildcards, to_mesh=False):
    # there are three options:
    # 1) no spatial dge
    # 2) spatial dge, no mesh
    # 3) spatial dge with a mesh
    run_mode = list(
        get_run_modes_from_sample(wildcards.project_id, wildcards.sample_id).keys()
    )[0]

    puck_barcode_file_ids = project_df.get_puck_barcode_ids_and_files(
        wildcards.project_id, wildcards.sample_id
    )[0]

    return [
        get_dge_from_run_mode(
            project_id=wildcards.project_id,
            sample_id=wildcards.sample_id,
            run_mode=run_mode,
            data_root_type=wildcards.data_root_type,
            downsampling_percentage=wildcards.downsampling_percentage,
            puck_barcode_file_id=puck_barcode_file_ids,
            only_spatial=True,
            to_mesh=to_mesh,
        )["dge"]
    ][0]


def get_barcode_files_matching_summary_input(wildcards):
    pbf_ids, pbfs = project_df.get_puck_barcode_ids_and_files(
        project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )

    parsed_spatial_barcode_files = [
        expand(
            parsed_spatial_barcodes,
            project_id=wildcards.project_id,
            sample_id=wildcards.sample_id,
            puck_barcode_file_id=pbf_id,
        )[0]
        for pbf_id in pbf_ids
    ]

    return {
        "puck_barcode_files": pbfs,
        "parsed_spatial_barcode_files": parsed_spatial_barcode_files,
    }


def get_barcode_summary_files_matching_summary_input(wildcards):
    pbf_ids, _ = project_df.get_puck_barcode_ids_and_files(
        project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )

    parsed_spatial_barcode_summary_files = [
        expand(
            parsed_spatial_barcodes_summary,
            project_id=wildcards.project_id,
            sample_id=wildcards.sample_id,
            puck_barcode_file_id=pbf_id,
        )[0]
        for pbf_id in pbf_ids
    ]

    return {"matched_barcode_files_summary": parsed_spatial_barcode_summary_files}


def get_barcode_files(wildcards):
    pbf_ids, pbfs = project_df.get_puck_barcode_ids_and_files(
        project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )

    return {"puck_barcode_files": pbfs}


def get_stats_prealigned_spatial_barcodes(wildcards):
    pbf_ids, pbfs = project_df.get_puck_barcode_ids_and_files(
        project_id=wildcards.project_id, sample_id=wildcards.sample_id
    )

    parsed_stats_prealigned_sp_bc = [
        expand(
            stats_prealigned_spatial_barcodes,
            project_id=wildcards.project_id,
            sample_id=wildcards.sample_id,
            puck_barcode_file_id=pbf_id,
        )[0]
        for pbf_id in pbf_ids
    ]

    return {
        "puck_barcode_files": pbfs,
        "stats_prealigned_spatial_barcodes": parsed_stats_prealigned_sp_bc,
    }


def get_automated_analysis_dge_input(wildcards):
    # there are three options:
    # 1) no spatial dge
    # 2) spatial dge, no mesh
    # 3) spatial dge with a mesh
    if wildcards.puck_barcode_file_id_qc != "puck_collection":
        return [
            get_dge_from_run_mode(
                project_id=wildcards.project_id,
                sample_id=wildcards.sample_id,
                run_mode=wildcards.run_mode,
                data_root_type=wildcards.data_root_type,
                downsampling_percentage=wildcards.downsampling_percentage,
                puck_barcode_file_id=wildcards.puck_barcode_file_id_qc,
            )["dge"]
        ]
    else:
        return [
            get_dge_collection_from_run_mode(
                project_id=wildcards.project_id,
                sample_id=wildcards.sample_id,
                run_mode=wildcards.run_mode,
                data_root_type=wildcards.data_root_type,
                downsampling_percentage=wildcards.downsampling_percentage,
            )["dge"]
        ]


def get_novosparc_input_files(config):
    if (
        "reference_project_id" in config
        and config["reference_project_id"] != ""
        and "reference_sample_id" in config
        and config["reference_sample_id"] != ""
        and "reference_run_mode" in config
        and config["reference_run_mode"] != ""
        and "reference_umi_cutoff" in config
        and config["reference_umi_cutoff"] != ""
    ):

        ret = expand(
            novosparc_with_reference_h5ad,
            data_root_type="complete_data",
            downsampling_percentage="",
            project_id=config["project_id"],
            sample_id=config["sample_id"],
            run_mode=config["run_mode"],
            umi_cutoff=config["umi_cutoff"],
            reference_project_id=config["reference_project_id"],
            reference_sample_id=config["reference_sample_id"],
            reference_run_mode=config["reference_run_mode"],
            reference_umi_cutoff=config["reference_umi_cutoff"],
        )

    elif (
        "project_id" in config
        and "sample_id" in config
        and "umi_cutoff" in config
        and "run_mode" in config
    ):
        ret = expand(
            novosparc_denovo_h5ad,
            data_root_type="complete_data",
            downsampling_percentage="",
            project_id=config["project_id"],
            sample_id=config["sample_id"],
            run_mode=config["run_mode"],
            umi_cutoff=config["umi_cutoff"],
        )
    else:
        ret = []

    return ret


def get_novosparc_with_reference_input_files(wildcards):
    out_dict = {}

    st_adata = expand(
        automated_analysis_result_file,
        data_root_type=wildcards.data_root_type,
        downsampling_percentage=wildcards.downsampling_percentage,
        # access data without the first underscore
        project_id=wildcards.reference_project_id,
        sample_id=wildcards.reference_sample_id,
        run_mode=wildcards.reference_run_mode,
        umi_cutoff=wildcards.reference_umi_cutoff,
    )

    out_dict["st_adata"] = st_adata

    return out_dict



__author__ = ["Marvin Jens"]
__email__ = ["marvin.jens@mdc-berlin.de"]

import os
import argparse
import logging
import pandas as pd
import numpy as np
from spacemake.util import ensure_path
import spacemake.longread.report as report
import spacemake.longread.cache as cache
import spacemake.longread.annotation as ann
from spacemake.longread.signature import (
    get_signature_db,
    process_intact_signature,
    digest_signatures,
)
from collections import defaultdict


def detect_sample(args):
    if args.sample is None:
        if not hasattr(args, "fname"):
            raise ValueError("please provide a sample-identifier via --sample")

        sample_name, _ = os.path.splitext(os.path.basename(args.fname))
        logging.info(f"auto-detected sample_name={sample_name}")
    else:
        sample_name = args.sample
    return sample_name.replace(".stats", "")


def store_results(df, path, fname, logger, **kw):
    fpath = os.path.join(ensure_path(path), fname)
    df.to_csv(fpath, sep="\t", **kw)
    logger.info(f"storing {len(df)} rows to '{fpath}'")
    return fpath


def load_results(path, fname, logger, **kw):
    fpath = os.path.join(path, fname)
    df = pd.read_csv(fpath, sep="\t", comment="#", **kw).fillna("other")
    logger.info(f"loaded {len(df)} rows from '{fpath}'")
    return df


def setup_namespace(args, need_sample_name=True):
    """
    perform initialization code common to more than one sub-command.
    Augments the args object returned from argparse module with additional
    variables that are used by multiple sub-commands.

    :param args: cmdline parser args
    :type args: Namespace
    :return: augmented namespace
    :rtype: SimpleNamespace
    """
    d = vars(args)
    if need_sample_name:
        sample = detect_sample(args)
        d["sample_name"] = sample

    db = get_signature_db(args.config)

    d["signature_db"] = db
    d["CB"] = db.CB[args.signature]
    d["UMI"] = db.UMI[args.signature]
    d["read1_primer"] = db.read1_primer[args.signature]
    d["read2_primer"] = db.read2_primer[args.signature]
    d["cDNA_after"] = db.cDNA_after[args.signature]
    intact_bead = db.intact[args.signature]
    d["intact_bead"] = intact_bead
    d["relevant"] = (
        intact_bead.split(",")
        + db.other[args.signature].split(",")
        + db.prefixes[args.signature].split(",")
        + db.suffixes[args.signature].split(",")
    )
    d["sig_intact"] = intact_bead.split(",")
    d["sig_core"], d["sig_core_order"] = process_intact_signature(intact_bead)
    d["bead_related"] = d["sig_intact"][0]
    d["logger"] = logging.getLogger(f"spacemake.longread.cmd.{args.subcmd}")
    d["blocks"] = db.blocks

    from types import SimpleNamespace

    return SimpleNamespace(**d)


def initialize(args):
    args = setup_namespace(args)

    annotation = ann.AnnotatedSequences(
        args.fname,
        os.path.join(args.annotation_out, f"{args.sample_name}.annotation.tsv"),
        args.sample_name,
        args.blocks,
        min_score=args.min_score,
        relevant=args.relevant,
        orient_by=args.bead_related,
    )
    n_total = len(annotation.raw_sequences)

    args.logger.info(
        f"total number of reads in {args.fname} ({args.sample_name}) is {n_total}"
    )
    return args, annotation


## Subcommand 'align'
def cmd_align(args):
    args = setup_namespace(args)

    cache.fill_caches(
        args.fname,
        args.sample_name,
        args.blocks,
        relevant=args.relevant,
        path=ensure_path(args.cache),
        n_proc=args.parallel,
    )

    df = cache.annotate(
        args.fname,
        args.sample_name,
        args.blocks,
        path=ensure_path(args.cache),
        relevant=args.relevant,
    )
    df.to_csv(
        os.path.join(
            ensure_path(args.annotation_out), f"{args.sample_name}.annotation.tsv"
        ),
        sep="\t",
        index=False,
    )


def prepare_align_parser(subparsers):
    parser = subparsers.add_parser("align", help="align PacBio reads against oligos")
    parser.add_argument(
        "fname",
        default=None,
        help="file with pacbio reads (FASTQ or BAM format)",
    )
    parser.set_defaults(func=cmd_align)
    return parser


## subcommand 'annotate'
def cmd_annotate(args):
    args, annotation = initialize(args)
    n_total = len(annotation.raw_sequences)

    sig_counts, n_concat, n_reprimed = annotation.count_signatures()
    report.count_dict_out(sig_counts, "common signatures", total=n_total)
    print(
        f"n_concat={n_concat} ({100.0 * n_concat/n_total:.2f}%) "
        f"n_reprimed={n_reprimed} ({100.0 * n_reprimed/n_total:.2f}%)"
    )
    df_sig = report.count_dict_to_df(sig_counts, "signatures", n_total=n_total)

    concat_counts, n_occurrences = annotation.count_concatenations()
    report.count_dict_out(concat_counts, "concatenations", total=n_occurrences)
    df_concat = report.count_dict_to_df(
        concat_counts, "concatenations", n_total=n_occurrences
    )

    reprimed_counts = annotation.count_repriming()
    report.count_dict_out(reprimed_counts, "repriming", total=n_total)
    df_reprimed = report.count_dict_to_df(reprimed_counts, "repriming", n_total=n_total)

    partial_counts, prefixes, suffixes, pT_counts = annotation.completeness(
        args.sig_core, polyT=args.polyT
    )

    partial_counts_simple, _ = report.count_dict_collapse_misc(
        partial_counts, sig_intact=args.sig_intact, total=n_total, misc_thresh=0.00001
    )
    report.count_dict_out(partial_counts_simple, "completeness", total=n_total)

    df_comp = report.count_dict_to_df(
        partial_counts_simple, kind="bead_complete", n_total=n_total
    ).sort_values("name")

    df_pT = report.count_dict_to_df(pT_counts, kind="polyT_after", n_total=n_total)

    df = pd.concat([df_sig, df_concat, df_reprimed, df_comp, df_pT])
    store_results(df, args.stats_out, f"{args.sample_name}.stats.tsv", args.logger)

    # TODO: prefix/suffix counts add up to > 100%. Needs fix
    report.count_dict_out(pT_counts, "polyT after", total=n_total)
    report.count_dict_out(prefixes, "prefixes", total=n_total)
    report.count_dict_out(suffixes, "suffixes", total=n_total)

    # Gather statistics about the parts that make up intact oligos
    qintact, qL, qstarts, qends, qscores = annotation.query_dimensions(
        args.sig_core, substring=True
    )
    # print(qstarts.shape, qintact.shape, qL.shape)
    from collections import Counter

    data = []
    for part, starts, ends, scores in zip(args.sig_core, qstarts.T, qends.T, qscores.T):
        starts_hist = sorted(Counter(starts).items())
        ends_hist = sorted(Counter(ends).items())
        lens_hist = sorted(Counter(ends - starts).items())
        scores_hist = sorted(Counter(scores).items())

        for x, f in starts_hist:
            data.append((args.intact_bead, part, "start", x, f))

        for x, f in ends_hist:
            data.append((args.intact_bead, part, "end", x, f))

        for x, f in lens_hist:
            data.append((args.intact_bead, part, "len", x, f))

        for x, f in scores_hist:
            data.append((args.intact_bead, part, "score", x, f))

    # For reference, also gather statistics for each part of the intact signature,
    # regardless of whether it occurs in the context of an intact signature or not,
    # i.e. what do *all matches* look like
    for part in args.sig_intact:
        qnames, starts, ends, scores, qL = annotation.query_oligo_occurrences(part)
        starts_hist = sorted(Counter(starts).items())
        ends_hist = sorted(Counter(ends).items())
        lens_hist = sorted(Counter(ends - starts).items())
        scores_hist = sorted(Counter(scores).items())

        for x, f in starts_hist:
            data.append(("anywhere", part, "start", x, f))

        for x, f in ends_hist:
            data.append(("anywhere", part, "end", x, f))

        for x, f in lens_hist:
            data.append(("anywhere", part, "len", x, f))

        for x, f in scores_hist:
            data.append(("anywhere", part, "score", x, f))

    df_parts = pd.DataFrame(
        data, columns=["signature", "oligo", "attr", "value", "freq"]
    )
    store_results(
        df_parts, args.stats_out, f"{args.sample_name}.intact_parts.tsv", args.logger
    )

    # output representative examples
    eo_fname = os.path.join(ensure_path(args.examples_out), f"{args.sample_name}.txt")

    with open(eo_fname, "wt") as eo:
        for signame, sigcount in sorted(sig_counts.items(), key=lambda x: -x[1]):
            i = None
            for i, (qname, _, _) in enumerate(
                annotation.filter_signatures(tuple(signame.split(",")))
            ):
                eo.write(
                    f"# example={i} {signame} n={sigcount}\n{annotation.fmt(qname)}\n"
                )
                if i >= 10:
                    break

            if i is None:
                args.logger.warning(
                    f"unable to find any reads with signature {signame}"
                )


def prepare_annotate_parser(subparsers):
    parser = subparsers.add_parser(
        "annotate", help="create annotation from detected oligo matches"
    )
    parser.add_argument(
        "fname",
        default=None,
        help="file with pacbio reads (FASTQ or BAM format)",
    )
    parser.add_argument(
        "--polyT",
        help="name of oligo-block that matches polyT (default='polyT')",
        default="polyT",
    )
    parser.set_defaults(func=cmd_annotate)
    return parser


## subcommand 'report'
def get_synth_rates(found_part_counts, all_parts, n_total):
    rates = []

    n0 = n_total
    x = range(1, len(all_parts) + 1)
    for i in x:
        key = tuple(all_parts[:i])
        # print(found_part_counts)
        # print(key, found_part_counts[key], n0, rates)
        rates.append(found_part_counts[key] / float(max(1, n0)))
        # print(i, rates)
        n0 = found_part_counts[key]

    return np.array(rates)


def cmd_report(args):
    args = setup_namespace(args)
    args.logger.info(f"generating report plots for '{args.sample_name}'")
    args.logger.debug(f"'bead-related' if we detect: '{args.bead_related}'")

    df = load_results(args.stats_out, f"{args.sample_name}.stats.tsv", args.logger)
    sig_counts = report.count_dict_from_df(df, "signatures")
    report.count_dict_out(
        sig_counts, "signatures", misc_thresh=0.05, total=sig_counts["n_total"]
    )
    n_total = sig_counts["n_total"]
    args.logger.info(f"n_total={n_total}")

    # disentangle bead-related from other signatures
    ov_counts, bead_counts, found_part_counts, core_signature = digest_signatures(
        sig_counts,
        args.bead_related,
        args.intact_bead,
        # prefixes=args.prefixes,
        # suffixes=args.suffixes,
    )
    # group low abundance signatures into 'misc' for overview donut plot
    ov_simple, _ = report.count_dict_out(
        ov_counts, "overview counts", misc_thresh=0.01, total=sig_counts["n_total"]
    )
    # group low abundance signatures into 'misc' for bead completeness donut plot
    bead_simple, _ = report.count_dict_out(
        bead_counts, "bead counts", misc_thresh=0.0, total=ov_counts["bead-related"]
    )
    # store the donut plot values in ...report.tsv for cross-sample overviews and such
    df_ov = report.count_dict_to_df(ov_simple, kind="overview")
    df_bead = report.count_dict_to_df(bead_simple, kind="bead_related")
    df_rep = pd.concat([df_ov, df_bead])
    df_rep["sample"] = args.sample_name
    df_rep["signature"] = args.signature
    print(df_rep)
    store_results(df_rep, args.stats_out, f"{args.sample_name}.report.tsv", args.logger)

    # compute and store apparent synthesis rates
    syn_rates = get_synth_rates(found_part_counts, core_signature, n_total)
    df = pd.DataFrame(dict(segment=core_signature, rate=syn_rates))
    df["sample"] = args.sample_name
    store_results(df, args.stats_out, f"{args.sample_name}.synth.tsv", args.logger)

    # prepare for plotting
    del ov_simple["n_total"]
    # assert np.array(list(ov_simple.values())).sum() == n_total
    print(
        f"sum of ov_simple {np.array(list(ov_simple.values())).sum()} n_total={n_total}"
    )

    ov_items = sorted(ov_simple.items(), key=lambda x: -x[1])
    ov_labels = [x[0] for x in ov_items]
    ov_counts = [x[1] for x in ov_items]

    bead_items = sorted(bead_simple.items(), key=lambda x: -x[1])
    bead_labels = [x[0] for x in bead_items]
    bead_counts = [x[1] for x in bead_items]

    # render the donut plots #
    report.plot_results(
        sig_counts,
        ov_labels,
        ov_counts,
        bead_labels,
        bead_counts,
        syn_rates,
        all_parts=core_signature,
        fname=os.path.join(
            ensure_path(args.report_out), f"{args.sample_name}.donuts.pdf"
        ),
        suptitle=args.sample_name,
    )

    # render the pre-computed position/score/length histograms as a plot
    report.plot_histograms(
        df=load_results(
            args.stats_out, f"{args.sample_name}.intact_parts.tsv", args.logger
        ),
        fname=os.path.join(
            ensure_path(args.report_out), f"{args.sample_name}.hists.pdf"
        ),
        n_total=n_total,
        parts=args.intact_bead.split(","),
    )


def prepare_report_parser(subparsers):
    parser = subparsers.add_parser("report", help="create PDF/PNG reports")
    parser.set_defaults(func=cmd_report)
    return parser


## subcommand 'edits'
def cmd_edits(args):
    args, annotation = initialize(args)

    data = []
    for part in args.sig_intact:
        qmatches = annotation.query_oligo_occurrences(part)
        if len(qmatches[0]) > args.n_samples:
            qmatches = ann.subsample(qmatches, n=args.n_samples)

        nmatch = len(qmatches[0])
        m, ed = ann.align_stats(
            annotation, args.blocks[part], qmatches, min_score=args.min_score
        )
        for x in np.arange(len(m)):
            # print(part, x, m)
            data.append((part, args.blocks[part], nmatch, x, m[x], ed[x]))

    df = pd.DataFrame(
        data, columns=["oligo", "seq", "nmatch", "pos", "fmatch", "ed_dict"]
    )
    store_results(
        df, args.stats_out, f"{args.sample_name}.oligo_edits.tsv", args.logger
    )

    report.plot_edits(
        df,
        os.path.join(args.report_out, f"{args.sample_name}.oligo_edits.pdf"),
        parts=args.sig_intact,
    )


def prepare_edits_parser(subparsers):
    parser = subparsers.add_parser(
        "edits", help="gather mismatch and indel stats for oligo matches"
    )
    parser.add_argument(
        "--n-samples",
        type=int,
        default=1000,
        help="number of sample alignments to gather edit statistics from (default=1000)",
    )
    parser.add_argument(
        "fname",
        default=None,
        help="file with pacbio reads (FASTQ or BAM format)",
    )
    parser.add_argument(
        "--min-score",
        default=0.6,
        type=float,
        help="minimal match alignment score to consider a match for edit extraction, relative to its size (default=0.6)",
    )
    parser.set_defaults(func=cmd_edits)
    return parser


## subcommand 'extract'
def cmd_extract(args):
    args, annotation = initialize(args)

    # anchor_scores = defaultdict(float)
    CB_detect = set()
    # umis = defaultdict(lambda: "NA")

    args.logger.info(
        "expecting read1 primed by {args.read1_primer} and CB={args.CB} UMI={args.UMI}"
    )
    args.logger.info(
        f"expecting cDNA after {args.cDNA_after} and primed by {args.read2_primer}"
    )

    if args.top_barcodes:
        known = set([bc.strip()[::-1] for bc in open(args.top_barcodes).readlines()])
        rev = set([bc[::-1] for bc in known])
        logger.info(f"loaded {len(known)} barcodes from '{args.top_barcodes}'. ")
    else:
        known = set()

    n = 0
    for qname, sig in annotation.signatures.items():
        sig_str = ",".join(sig)
        if args.sig_include and not (args.sig_include in sig_str):
            continue

        if args.sig_exclude and (args.sig_exclude in sig_str):
            continue

        r1 = ""
        if args.read1_primer in sig:
            ((r1_start, r1_end), (start_oli, end_oli), r1,) = annotation.extract_cDNA(
                qname, after_oligo=args.read1_primer, distal=args.distal
            )

        r2 = ""
        if args.cDNA_after in sig:
            (
                (cDNA_start, cDNA_end),
                (start_oli, end_oli),
                r2,
            ) = annotation.extract_cDNA(
                qname, after_oligo=args.cDNA_after, distal=args.distal
            )
            if not r2.strip() or (cDNA_end - cDNA_start <= 0):
                r2 = ""

        CB = eval(args.CB, dict(r1=r1)) if r1 else "NA"
        UMI = eval(args.UMI, dict(r1=r1, r2=r2)) if r1 and r2 else "NA"

        CB_detect.add(CB)
        if CB not in known:
            CB = CB.lower()

        n += 1
        seq = annotation.raw_sequences[qname]
        if r2:
            header = (
                f">{n}__CB:{CB}__UMI:{UMI}__"
                + f"sig:{sig_str}__cDNA:{cDNA_start}-{cDNA_end}__oli:{start_oli}-{end_oli}__"
                + f"L_read={len(seq)}__L_cDNA={cDNA_end - cDNA_start}"
            )
            # print(header[: (254 - 14)] + " 1:N:0:TCCTGAGC" + f"\n{cDNA}")
            print(header[: (254 - 14)] + " 1:N:0:TCCTGAGC" + f"\n{r2}")

    if len(known):
        nd = len(CB_detect)
        ov = len(CB_detect & known)
        ctrl = len(CB_detect & rev)
        args.logger.info(
            f"{ov} / {nd} detected barcodes overlap ({ov/nd * 100:.3f}%). "
            f"Reverse BC control {ctrl/nd * 100:.3f}%"
        )


def prepare_extract_parser(subparsers):
    parser = subparsers.add_parser(
        "extract", help="extract sequences from the long read"
    )
    parser.add_argument(
        "fname",
        default=None,
        help="file with pacbio reads (FASTQ or BAM format)",
    )
    parser.add_argument(
        "--barcode-after",
        type=str,
        default="bead_start",
        help="name of anchor match after which the barcodes follow (default='bead_start')",
    )
    parser.add_argument(
        "--cDNA-after",
        type=str,
        default="bead_start",
        help="excise sequence between this oligo and the last oligo match (if it exists)",
    )
    parser.add_argument(
        "--sig-include",
        type=str,
        default="",
        help="extract cDNA only from long reads INCLUDING this substring in the signature",
    )
    parser.add_argument(
        "--sig-exclude",
        type=str,
        default="",
        help="extract cDNA only from long reads EXCLUDING this substring in the signature",
    )
    parser.add_argument(
        "--distal",
        default=150,
        type=int,
        help="number of nt from end of sequence that are considered for oligo matches when extracting cDNA",
    )
    parser.add_argument(
        "--top-barcodes",
        type=str,
        default="",
        help="path to text file with known barcodes (e.g. from Illumina)",
    )
    parser.set_defaults(func=cmd_extract)
    return parser


## subcommand 'overview'
def cmd_overview(args):
    args = setup_namespace(args, need_sample_name=False)

    dfs = []
    for fname in list(args.reports):
        print(f"loading {fname}")
        df = pd.read_csv(fname, sep="\t", index_col=0)
        dfs.append(df)

    df = pd.concat(dfs, ignore_index=True)

    dfsig = df[["sample", "signature"]].drop_duplicates().set_index("sample")
    print(dfsig)
    df = df[["name", "count", "sample"]].pivot(
        index="sample", columns="name", values="count"
    )  # .query("kind == 'overview'")
    df = df.fillna(0).div(df["n_total"], axis=0)
    df *= 100  # we'd like percentages

    repriming = [
        "TSO,TSO_RC",
        "dN-SMRT,dN-SMRT_RC",
    ]
    concatenation = [c for c in df.columns if "+" in c]
    bead = ["complete", "only_bead_start", "missing_polyT", "missing_OP1"][::-1]

    # # avoid crash if columns are missing
    for r in repriming + concatenation + bead:
        if r not in df.columns:
            df[r] = 0

    # print(df)
    # print(f"concat columns {concatenation}")
    # print(f"bead columns {bead}")
    df["reprimed"] = df[repriming].sum(axis=1)
    # df["complete"] = np.nan_to_num(df["complete"], nan=0.0)
    df["concat"] = df[concatenation].sum(axis=1)
    df["bead_related"] = np.nan_to_num(df[bead].sum(axis=1), nan=0.0)

    def ds(row):
        sig = dfsig.loc[row.name]
        res = (sig == "dropseq") * row.complete
        return res

    df["dropseq"] = df[["complete"]].apply(ds, axis=1)
    # df["complete"] = df[""]
    df["incomplete"] = df["bead-related"] - df["complete"] - df["dropseq"]
    df["non_bead"] = 100 - df["bead-related"]
    df["fidelity"] = 100 * df["complete"] / df["bead-related"]
    df = df.fillna(0)
    # print(df)

    df = df.reset_index().sort_values("sample")
    store_results(
        df,
        args.output,
        "overview.csv",
        float_format="%.2f",
        logger=args.logger,
    )
    report.overview_plots(
        df,
        path=args.output,
        bead=bead,
        concatenation=concatenation,
        repriming=repriming,
    )

    return df


def prepare_overview_parser(subparsers):
    parser = subparsers.add_parser(
        "overview", help="make overview plots across samples"
    )
    parser.add_argument("reports", nargs="+", help="sample reports to aggregate")
    parser.add_argument(
        "--output",
        default=".",
        help="path for detailed report PDFs and CSV overview table",
    )
    parser.set_defaults(func=cmd_overview)
    return parser


## main commandline interface
def prepare_toplevel_parser():
    parser = argparse.ArgumentParser(prog="longread")
    parser.add_argument(
        "--debug", default=False, action="store_true", help="activate debug output"
    )
    parser.add_argument(
        "--sample",
        help="sample name (default=autodetect from fname)",
        default=None,
    )
    parser.add_argument(
        "--config",
        help="YAML describing expected longread signature (default=./longread.yaml if detected or built-in)",
        default="longread.yaml",
    )
    parser.add_argument(
        "--signature",
        help="expected long read signature (e.g. dropseq, chromium, noUMI, withUMI,...)",
        default=None,
    )
    parser.add_argument(
        "--blocks",
        default="",
        help="FASTA file with known oligo sequences (default=built-in)",
    )
    parser.add_argument(
        "--cache",
        default="./cache/",
        help="path to alignment caches (default=./cache/)",
    )
    parser.add_argument(
        "--annotation-out",
        default="./annotation/",
        help="path to store annotation data in (default=./annotation/)",
    )
    parser.add_argument(
        "--examples-out",
        help="path to store annotated example read sequences",
        default="./examples/",
    )
    parser.add_argument(
        "--stats-out",
        help="path to store statistics (pandas dataframes)",
        default="./stats/",
    )
    parser.add_argument(
        "--report-out", help="path to render graphical reports in", default="./reports/"
    )
    parser.add_argument(
        "--parallel",
        default=16,
        type=int,
        help="number of parallel processes (default=16)",
    )
    parser.add_argument(
        "--min-score",
        default=0.6,
        type=float,
        help="minimal match alignment score to consider a match for annotation, relative to its size (default=0.6)",
    )

    ## sub-parser setup ##
    sp = parser.add_subparsers(help="sub-command help", dest="subcmd")
    prepare_align_parser(sp)
    prepare_annotate_parser(sp)
    prepare_edits_parser(sp)
    prepare_extract_parser(sp)
    prepare_report_parser(sp)
    prepare_overview_parser(sp)

    parser.set_defaults(func=lambda args: parser.print_help())

    return parser


def cmdline():
    """
    Commandline interface to handle long-read (PacBio, nanopore, ...)
    annotation and analysis.

    :return: depends on sub-command used
    :rtype: depends on sub-command used
    """
    import logging

    parser = prepare_toplevel_parser()
    args = parser.parse_args()
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    return args.func(args)


if __name__ == "__main__":
    result = cmdline()
